try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
from OpenGL import GL
import numpy as np
from .VisualizationBaseObject import VisualizationBaseObject, config
from .Shaders import Shader
from .Tools import glm

_cs_vs = files('phybers.fibervis') / 'shaders' / 'coordinateSystem.vs'
_sfs_fs = files('phybers.fibervis') / 'shaders' / 'standardFragmentShader.fs'

class CoordinateSystem(VisualizationBaseObject):
    def __init__(self, shaderProgram):

        super().__init__()

        self.shader = shaderProgram

        radiusCilinder = 0.015
        lengthCilinder = 0.8

        radiusCone = 0.05
        lengthCone = 0.2

        self.points, self.elements = CoordinateSystem.createArrow(radiusCilinder, lengthCilinder, radiusCone, lengthCone, detail=10)

        model0 = glm.Quaternion.identity().rotation4Matrix()
        model1 = glm.Quaternion.fromAngleAxis(np.pi/2, (0,0,1)).rotation4Matrix()
        model2 = glm.Quaternion.fromAngleAxis(-np.pi/2, (0,1,0)).rotation4Matrix()

        self.model = np.concatenate(( model0.getA(), model1.getA(), model2.getA() ))

        self.color = np.array([1,0,0,1, 0,1,0,1, 0,0,1,1], dtype=np.float32)

        self._loadBuffers()


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(1, [self.vao])

        GL.glDeleteBuffers(1, [self.vbo])
        GL.glDeleteBuffers(1, [self.ebo])

        self.clean = True

    @staticmethod
    def createArrow(rCilinder, lCilinder, rCone, lCone, detail=3):
        # Creates an arrow on the x-axis, with the specified measure.
        # detail is the number of faces, 3 = pyramid
        #
        # Parameters
        # ----------
        # rCilinder : float
        #       radius for the cilinder.
        # lCilinder : float
        #       lengh for the arrow before the head.
        # rCone : float
        #       radius for the arrow head.
        # lCone : float
        #       length of the arrow head.
        # detail : int
        #       number of faces to build the arrow.
        #
        # Returns
        # -------
        # arrowPoints : numpy.array
        #       An array with the points.
        # elements : numpy.array
        #       Array with the order of the drawing.

        if not isinstance(detail, int):
            raise TypeError('Variable detail must be a integer major than 2.')

        n = (detail+1)*3 *3
        arrowPoints = np.zeros(n, dtype=np.float32)

        arrowPoints[3] = lCilinder
        arrowPoints[6] = lCilinder+lCone

        angle = 2.0*np.pi/detail
        line = [0.0,            rCilinder,      0.0,
                lCilinder,      rCilinder,      0.0]

        for i in range(detail):
            rotation = glm.Quaternion.fromAngleAxis(angle*i, [1,0,0])
            arrowPoints[9+i*6 : 12+i*6] = rotation.rotateVector(line[:3])
            arrowPoints[12+i*6 : 15+i*6] = rotation.rotateVector(line[3:])

        k = 9 + detail*6
        line = [lCilinder, rCone, 0]
        for i in range(detail):
            rotation = glm.Quaternion.fromAngleAxis(angle*i, [1,0,0])
            arrowPoints[k+i*3 : k+i*3+3] = rotation.rotateVector(line)

        elements = []

        for i in range(detail):
            elements.append(0)
            elements.append(i*2+3)
            elements.append(i*2+2+3)
        elements[-1] = 3

        for i in range(detail):
            elements.append(3+i*2)
            elements.append(3+i*2+1)
            elements.append(3+i*2+2)

            elements.append(3+i*2+1)
            elements.append(3+i*2+2)
            elements.append(3+i*2+3)
        elements[-1] = 4
        elements[-2] = 3
        elements[-4] = 3

        for i in range(detail):
            elements.append(1)
            elements.append(i + 3+2*detail)
            elements.append(i+1+3+2*detail)
        elements[-1] = 3+2*detail

        for i in range(detail):
            elements.append(2)
            elements.append(i + 3+2*detail)
            elements.append(i+1+3+2*detail)
        elements[-1] = 3+2*detail

        elements = np.array(elements, np.int32)

        return arrowPoints, elements


    def _loadBuffers(self):
        ''' It generates a VAO, VBO and EBO.
        It populates them with the vertex (the VBO), also the elements (the EBO).

        It finishes when the buffers are linked with attributes on the vertex shader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        # Reference for vbos, ebos and attribute ptrs
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        self.vbo = GL.glGenBuffers(1)
        self.ebo = GL.glGenBuffers(1)

        # VBO
        # bufferSize = self.points.nbytes
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)
        GL.glBufferData(GL.GL_ARRAY_BUFFER, self.points.nbytes, self.points.flatten(), GL.GL_STATIC_DRAW)       # Create empty buffer

        # Populate buffer with points
        # GL.glBufferSubData(GL.GL_ARRAY_BUFFER,
        #       0,
        #       self.points.nbytes,
        #       self.points.flatten())

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)
        GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, self.elements.nbytes, self.elements.flatten(), GL.GL_STATIC_DRAW)

        # Enable attributes
        positionAttribute =     self.shader[0].attributeLocation('vertexPos')

        # # Connect attributes
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute, 3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)


    def loadUniforms(self):

        GL.glUniform4fv(self.shader[self.selectedShader].glGetUniformLocation('colorArray'), 3, self.color)
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("M"), 3, GL.GL_TRUE, self.model)



    @config
    def draw(self):
        GL.glDrawElementsInstanced(GL.GL_TRIANGLES, len(self.elements), GL.GL_UNSIGNED_INT, None, 3)

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_cs_vs)]
        fragmentGLSL = [str(_sfs_fs)]
        return [Shader(vertexGLSL, fragmentGLSL)]



from .FiberVis import main, process_args, parser


if __name__ == "__main__":
    ns, args = parser.parse_known_args()
    if ns.help:
        parser.print_help()
    else:
        main(**process_args(ns))



# -- coding: utf-8 --
"""
HClust Sub-Module.
"""

import gc
import os
import time
import numpy as np
from shutil import rmtree
from . import clusterTools as CT
from tempfile import mkdtemp
from pathlib import Path
from ...utils import read_bundle, write_bundle
from .c_wrappers import fiberDistanceMax, getAffinityGraphFromDistanceMatrix, getAverageLinkHCFromGraphFile
from ...utils import sampling


def hierarchical(fiber_input, outfile_dir, MaxDistance_Threshold):
    """
    Run to Hierarchical clustering
    """
    MatrixDist_output = os.path.join(outfile_dir, 'matrixd.bin')
    affinities_graph_output = os.path.join(outfile_dir,'affin.txt')
    dendogram_output = os.path.join(outfile_dir,'dendogram.txt')

    # Step1. Distance Matrix
    t0 = time.time()
    fiberDistanceMax(fiber_input, MatrixDist_output)
    gc.collect()
    print("Distance Matrix Delay: ", time.time()-t0, "[s]")

    # Step2. Affinities Graph
    t0= time.time()
    getAffinityGraphFromDistanceMatrix(
        MatrixDist_output, affinities_graph_output, MaxDistance_Threshold)
    gc.collect()
    print("Affinities Graph Delay: ", time.time()-t0, "[s]")

    # Step3. Dendogram
    t0= time.time()
    getAverageLinkHCFromGraphFile(
        affinities_graph_output, dendogram_output)
    gc.collect()
    print("Dendogram Delay: ", time.time()-t0, "[s]")


def particional_hierarchical(fiber_input, outfile_dir, PartDistance_Threshold, var, final_bundles_dir):

    """Writes the cluster bundles (bundles format) and fiber indexes per cluster (file '.txt')
    """

    arbfile = os.path.join(outfile_dir, 'dendogram.txt')
    afffile = os.path.join(outfile_dir, 'affin.txt')
    partfile = os.path.join(outfile_dir, 'index_fiberscluster.txt')

    t0= time.time()
    wfv=CT.wforest_partition_maxdist_from_graph(arbfile,PartDistance_Threshold,True,afffile,var)

    clusteres=wfv.clusters

    ar=open(partfile,'wt')
    ar.write(str(clusteres))
    ar.close()

    tractography = np.array(read_bundle(fiber_input))

    os.makedirs(final_bundles_dir, exist_ok=True)

    for i, j in enumerate(clusteres):

        write_bundle(os.path.join(final_bundles_dir, f"{i}.bundles"), tractography[j])

    print("Particional Hierarchical Delay: ", time.time()-t0, "[s]")


def is_reversed(cluster):
    base = np.stack([cluster, cluster[:, ::-1]])
    x = base - cluster[None, 0]
    x = np.square(x)
    x = np.sum(x, axis=-1)
    x = np.max(x, axis=-1)
    return x[0] > x[1]

def cal_centroide(cluster):
    cluster = np.asarray(cluster)
    r = is_reversed(cluster)
    if np.any(r):
        c_copy = np.empty_like(cluster)
        c_copy[~r] = cluster[~r]
        c_copy[r] = cluster[r, ::-1]
        cluster = c_copy
    return np.sum(cluster, axis=0) / len(cluster)

def write_centroids(clusters_dir, file_out):
    p = Path(clusters_dir)
    centroids = []
    clusters_paths = sorted(p.glob("*.bundles"), key=lambda x: int(x.stem))
    for cluster_path in clusters_paths:
        cluster = read_bundle(str(cluster_path))
        centroids.append(cal_centroide(cluster))
    write_bundle(file_out, centroids)


def hclust(file_in: str, dir_out: str, fiber_thr: int, partition_thr: int, variance: int) -> None:
    """
    Average-link hierarchical agglomerative clustering algorithm which allows finding bundles based on a pairwise fiber distance measure.

    Parameters
    ----------
    file_in : str
        Tractography data file in *'.bundles'* format.
    dir_out : str
        Directory to store all the results generated by the algorithm.
    fiber_thr : str
        Maximum distance threshold in *mm*, default: *30*.
    partition_thr : str
        Partition threshold in *mm*, default: *40*.
    variance : str
        Variance squared and provides a similarity scale in *mm*, default: *3600*.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Clusters : bundles files
        Directory that stores all the fiber clusters found in different '.bundles' files.
        The file names are labeled with integer numbers ranging from zero to the total number of fiber clusters found.
    Centroids : bundles file
        Directory that contains the centroid for each created cluster in same *'.bundles'* files.
        The firt fiber of bundle centroid is corresponding with centroid calculated for cluster one and so on
    Index of fibers per clusters : text file
        Text file that stores the fiber index input for each of the detected clusters.
        The fiber indexes are extracted from the tractography input. The first line corresponds to cluster zero and so on.

    Examples
    --------
    To test `hclust()`,  download the data from the `link hclust  <https://www.dropbox.com/sh/dt196k65v03eh9m/AABKRW7ad2ssB0N_dpjqK4Dha?dl=1>`_.
    Then, open a Python terminal and run the following commands.

    >>> from phybers.clustering import hclust
    >>> hclust (file_in = 'fibers_test.bundles', dir_out = 'hclust_result', fiber_thr = 30, partition_thr = 40, variance = 3600)

    You will locate the clustering results in the 'hclust_result' directory.
    """
    if os.path.exists(dir_out):
        rmtree(dir_out)

    os.mkdir(dir_out)

    final_bundles_dir = os.path.join(dir_out, 'FinalBundles')
    os.makedirs(final_bundles_dir, exist_ok=True)

    final_centroids_dir = os.path.join(dir_out, 'FinalCentroids')
    final_centroids_file = os.path.join(final_centroids_dir, 'centroids.bundles')
    os.makedirs(final_centroids_dir, exist_ok=True)

    outfile_dir = os.path.join(dir_out, 'outfile')
    os.makedirs(outfile_dir, exist_ok=True)

    np = 0
    data = read_bundle(file_in)
    for i in range(len(data)-1):
        if len(data[i]) != len(data[i+1]):
            np = 21
            break

    if np == 21:

        final_bundles21p_dir = os.path.join(dir_out, 'FinalBundles21p')
        os.makedirs(final_centroids_dir, exist_ok=True)

        fibers21p = os.path.join(outfile_dir, 'fiberorig_21p.bundles')
        sampling(file_in, fibers21p, np)

        hierarchical(fibers21p, outfile_dir,
                                fiber_thr)
        particional_hierarchical(
            fibers21p, outfile_dir, partition_thr, variance, final_bundles21p_dir)
        write_centroids(final_bundles_dir, final_centroids_file)

    else:
        hierarchical(file_in, outfile_dir, fiber_thr)
        particional_hierarchical(
            file_in, outfile_dir, partition_thr, variance, final_bundles_dir)
        write_centroids(final_bundles_dir, final_centroids_file)



from OpenGL import GL
from PyQt5 import QtCore, QtWidgets
from numpy import pi, sin

from .Framework.Tools import glm
from .Framework.Tools.visualizationEnums import SegmentationTypes, ROIType, VisualizationActions, VisualizationObject
from .Framework.CoordinateSystem import CoordinateSystem
from .Framework.Tools.Camera import Camera

from .Framework.BoundingBox import BoundingBox
from .Framework.Bundle import Bundle
from .Framework.Mesh import Mesh
from .Framework.MRI import MRI
from .Framework.Slice import Slice
from .Framework.Volume import Volume

from .Framework.Segmentation.SegmentationHandler import SegmentationHandler
from .Framework.Segmentation.InPlaceSegmentation import InPlaceSegmentation
from .Framework.Segmentation.ROISegmentation import ROISegmentation
from .Framework.Segmentation.ROI import ROI
from .Framework.Segmentation.AtlasBasedParallelSegmentation import AtlasBasedParallelSegmentation
from .Framework.Segmentation.FFClustSegmentation import FFClustSegmentation

# Performance test
import numpy as np
import time as time


SCALE_FACTOR = 1/12000


def defaultLighting():
    return {
    "lightPos" : [0.0, 100.0, 0.0],
    "lightAttr" : [0.5, 0.6, 1.0],
    "materialAttr" : [1.0, 1.0, 1.0], # White background for the GUI
    "shininess" : 5.0
    }


def defaultBackgroundColor():
	return [1.0, 1.0, 1.0, 1.0]


class GLWidget(QtWidgets.QOpenGLWidget):
    ''' Widget with the logic of the program.
    It uses framework classes for visualization. And it manages all the
    framework instances.
    '''

    updateViewingObjects = QtCore.pyqtSignal()
    updateSettingWindow = QtCore.pyqtSignal(SegmentationHandler)
    objectsChanged = QtCore.pyqtSignal(list)

    def __init__(self, parent = None):
        ''' Initialize the members for the OpenGL Widget
        '''

        super(GLWidget, self).__init__(parent)

        # self.version_profile = version_profile

        self.bundles = []
        self.meshes = []
        self.mris = []

        self.rois = []

        self.clearColor = defaultBackgroundColor()

        # default mouse flags
        self.orbit = False
        self.pan = False
        self.panning = False
        self.tilt = False

        self.currentObject = None
        self.draw_boundingbox = True

        # Mouse-screen attri
        self.screenX, self.screenY = 0,0

        self.segmentations = {  SegmentationTypes.InPlace : InPlaceSegmentation,
                                                        SegmentationTypes.ROIs : ROISegmentation,
                                                        SegmentationTypes.AtlasBased : AtlasBasedParallelSegmentation,
                                                        SegmentationTypes.FFClust : FFClustSegmentation }

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu) # type: ignore

        self.customContextMenuRequested.connect(self.ShowContextMenu)


    def initializeGL(self):
        ''' Initialize the OpenGL context and creates all the shaders programs
        '''

        print("OpenGL version: {}".format(GL.glGetString(GL.GL_VERSION).decode("utf-8"))) # type: ignore

        # Initialize shaders
        self.shaderDict = {}
        self.shaderList = [     Bundle.createProgram(),
                                                Mesh.createProgram(),
                                                Volume.createProgram(),
                                                BoundingBox.createProgram(),
                                                SegmentationHandler.createProgram(),
                                                # ROISegmentation.createProgram(),
                                                AtlasBasedParallelSegmentation.createProgram(),
                                                Slice.createProgram(),
                                                ROI.createProgram()]

        objectsList = [ [Bundle], #inplacesegmentation muved to segmentationhandler program
                                        [Mesh],
                                        [Volume],
                                        [BoundingBox],
                                        [InPlaceSegmentation, ROISegmentation],
                                        [AtlasBasedParallelSegmentation, FFClustSegmentation],
                                        [Slice],
                                        [ROI]]

        for i in range(len(objectsList)):
            self.shaderDict.update(dict.fromkeys(objectsList[i], self.shaderList[i]))

        self.coordSystemShader = CoordinateSystem.createProgram()

        self.camera = Camera(350.0)
        self.csCamera = Camera(1/sin(22.5*pi/180))

        # Config light parameters and view
        for s in self.shaderList:
            for shader in s:
                try:
                    self._configLight(shader, defaultLighting())
                except:
                    pass
                self._configView(shader)

        self._configView(self.coordSystemShader[0], self.csCamera)

        # Initialize some variables
        self.coordSystem = CoordinateSystem(self.coordSystemShader)
        self.proj = glm.perspective(45.0, 1, 0.01, 100000.0)
        self._configPerspective(self.coordSystemShader[0])


        self.testing = False
        self.FrameTime = GL.glGenQueries(1)

        ##########
        print(GL.GL_MAX_UNIFORM_BLOCK_SIZE, GL.glGetIntegerv(GL.GL_MAX_UNIFORM_BLOCK_SIZE))
        print(GL.GL_MAX_VERTEX_UNIFORM_COMPONENTS, GL.glGetIntegerv(GL.GL_MAX_VERTEX_UNIFORM_COMPONENTS))
        print(GL.GL_MAX_TEXTURE_SIZE, ' {} max dim.'.format(GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)))
        print(GL.GL_MAX_3D_TEXTURE_SIZE, ' {} max dim.'.format(GL.glGetIntegerv(GL.GL_MAX_3D_TEXTURE_SIZE)))



        GL.glEnable(GL.GL_DEPTH_TEST)
        GL.glEnable(GL.GL_PRIMITIVE_RESTART)
        GL.glPrimitiveRestartIndex(0xFFFFFFFF)


    def paintGL(self):
        ''' Paint function. Its called by the widget.
        '''

        GL.glViewport(0, 0, self.width(), self.height())

        GL.glClearColor(*self.clearColor)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        # FRAME TIME
        if self.testing:
            GL.glBeginQuery(GL.GL_TIME_ELAPSED, self.FrameTime)

        for bundle in self.bundles:
            bundle.draw()

        for mri in self.mris:
            mri.draw()

        for roi in self.rois:
            roi.draw()

        for mesh in self.meshes:
            mesh.draw(self.camera.eye)

        GL.glClear(GL.GL_DEPTH_BUFFER_BIT)
        GL.glViewport(0, 0, 150, 150)
        self.coordSystem.draw()

        # FRAME TIME
        if self.testing:
            GL.glEndQuery(GL.GL_TIME_ELAPSED)  # Termina consulta.
            ready = False
            while not ready:
                ready = GL.glGetQueryObjectuiv(self.FrameTime, GL.GL_QUERY_RESULT_AVAILABLE)

            self.testArray[self.testing_i] = (GL.glGetQueryObjectuiv(
                    self.FrameTime, GL.GL_QUERY_RESULT)/1000000000.0)
            self.testing_i += 1


    def resizeGL(self, width, height):
        ''' Recalculate the projection matrix, and upload the new one to all shaders
        '''

        self.proj = glm.perspective(45.0, width/height, 0.01, 100000.0)
        for s in self.shaderList:
            for shader in s:
                self._configPerspective(shader)


    def validBundleExtension(self):
        return Bundle.validExtension()


    def validMeshExtension(self):
        return Mesh.validExtension()


    def validMRIExtension(self):
        return MRI.validExtension()


    def addBundleFile(self, bundleFiles):
        # Makes the context the current one
        self.makeCurrent()

        new_bundles = [Bundle(bundlePath, self.shaderDict, self.bundles) for bundlePath in bundleFiles]
        if new_bundles:
            self.bundles.extend(new_bundles)

            if hasattr(self, 'camera'):
                bboxs = np.array([bundle.calculateBoundingBoxDimCenter() for bundle in new_bundles])
                center = (bboxs[:, 0].max(axis=0) + bboxs[:, 0].min(axis=0)) / 2

                self.camera.center = center
                self.camera.view = self.camera.calculateView()

                for s in self.shaderList:
                    for shader in s:
                        self._configView(shader)

            self._configView(self.coordSystemShader[0], self.csCamera)

        self.updateViewingObjects.emit()
        self.update()


    def addMeshFile(self, meshFiles):
        # Makes the context the current one
        self.makeCurrent()

        for meshPath in meshFiles:
            self.meshes.append(Mesh(meshPath, self.shaderDict, self.meshes))
            # self.meshes.append(Mesh(meshPath, self.meshShader, self.meshes))

        self.updateViewingObjects.emit()
        self.update()


    def addMRIFile(self, mriFiles):
        # Makes the context the current one
        self.makeCurrent()

        for mriPath in mriFiles:
            self.mris.append(MRI(mriPath, self.shaderDict, self.mris))

        self.updateViewingObjects.emit()
        self.update()


    def addROI(self):
        self.makeCurrent()

        self.rois.append(ROI(ROIType.Sphere, self.camera.center, 30.0, self.shaderDict, self.rois))

        self.updateViewingObjects.emit()
        self.update()


    @QtCore.pyqtSlot(dict)
    def modifyObject(self, options):
        # Makes the context the current one
        self.makeCurrent()

        def retrieveItemIteratively(root, index):
            item = root

            for i in index:
                if hasattr(item, 'children'):
                    item = item.children[i]
                else:
                    item = item[i]

            return item

        objectRoot = None
        itemObject = None

        rootType = options['rootType']
        index = options['number']
        action = options['action']

        if rootType == VisualizationObject.Bundle:
            objectRoot = self.bundles
        elif rootType == VisualizationObject.Mesh:
            objectRoot = self.meshes
        elif rootType == VisualizationObject.MRI:
            objectRoot = self.mris
        elif rootType == VisualizationObject.ROI:
            objectRoot = self.rois
        else:
            raise TypeError('Unrecognise visualization objects type: ', rootType)

        if index != -1:
            itemObject = retrieveItemIteratively(objectRoot, index)


        # Delete
        #########################
        if action == VisualizationActions.Delete:
            if index != -1:
                itemObject.cleanOpenGL()
                itemObject.parent.remove(itemObject)
                itemObject.clearRef()
            else:
                while(len(objectRoot) != 0):
                    item = objectRoot[0]
                    item.cleanOpenGL()
                    objectRoot.remove(item)
                    item.clearRef()
            self.updateViewingObjects.emit()

        # Toggle drawable
        #########################
        elif action == VisualizationActions.ToggleDrawable:
            itemObject.drawable = not itemObject.drawable

        # Rotate
        #########################
        elif action == VisualizationActions.Rotate:
            angle, axis = options['data']

            if index == -1:
                for i in objectRoot:
                    i.rotate(self.camera.center, angle, axis)
                self.objectsChanged.emit(objectRoot)

            else:
                itemObject.rotate(self.camera.center, angle, axis)
                self.objectsChanged.emit([itemObject])

        # Translate
        #########################
        elif action == VisualizationActions.Translate:
            v = options['data']

            if index == -1:
                for i in objectRoot:
                    i.translate(v)
                self.objectsChanged.emit(objectRoot)
            else:
                itemObject.translate(v)
                self.objectsChanged.emit([itemObject])

        # Scale
        #########################
        elif action == VisualizationActions.Scale:
            v = options['data']

            if index == -1:
                for i in objectRoot:
                    i.scale(v, self.camera.center)
                self.objectsChanged.emit(objectRoot)

            else:
                itemObject.scale(v, self.camera.center)

                self.objectsChanged.emit([itemObject])

        # ResetTransforms
        ##########################}
        elif action == VisualizationActions.ResetTransforms:
            if index == -1:
                for i in objectRoot:
                    i.resetModel()
                self.objectsChanged.emit(objectRoot)

            else:
                itemObject.resetModel()

                self.objectsChanged.emit([itemObject])

        # Load and Apply Matrix
        elif action == VisualizationActions.LoadAndApplyMatrix:
            matrixFile = options['data']

            itemObject.loadAndApplyMatrix(matrixFile)

        # Add Segmentation
        # ########################
        elif action == VisualizationActions.AddSegmentation:
            segmentationType = options['data']

            if segmentationType == SegmentationTypes.InPlace:
                segmentationObject = InPlaceSegmentation(itemObject, self.shaderDict)

            elif segmentationType == SegmentationTypes.ROIs:
                segmentationObject = ROISegmentation(itemObject, self.shaderDict)

            elif segmentationType == SegmentationTypes.AtlasBased:
                segmentationObject = AtlasBasedParallelSegmentation(itemObject, self.shaderDict)

            elif segmentationType == SegmentationTypes.FFClust:
                segmentationObject = FFClustSegmentation(itemObject, self.shaderDict)

            # Volume visualizations
            # #####################
            elif segmentationType == VisualizationObject.MRISlice:
                segmentationObject = Slice(itemObject, self.shaderDict)

            elif segmentationType == VisualizationObject.MRIVolume:
                segmentationObject = Volume(itemObject, self.shaderDict, self.camera)


            else:
                raise TypeError('Segmentation type not implemented yet: ', segmentationType)

            itemObject.children.append(segmentationObject)
            self.updateViewingObjects.emit()

        # Slice Modification
        elif action == VisualizationActions.SliceModification:
            linearInterp, discardValues, threshold, axis, pos2Axis, bright, contrast = options['data']

            itemObject.setLinearInterpolation(linearInterp)
            itemObject.setDiscardValuesWithThreshold(discardValues, threshold)
            itemObject.calculatePlaneNormal(axis, pos2Axis)
            itemObject.setBright(bright)
            itemObject.setContrast(contrast)

        # Mesh Modification
        elif action == VisualizationActions.MeshModification:
            if len(options['data']) == 1:
                if isinstance(options['data'][0], bool):
                    itemObject.setFront2BackSorting(options['data'][0])
                else:
                    itemObject.setColor(options['data'][0])

            else:
                drawLines, alpha = options['data']

                if index == -1:
                    for i in objectRoot:
                        i.setDrawLines(drawLines)
                        i.setAlpha(alpha)
                else:
                    itemObject.setDrawLines(drawLines)
                    itemObject.setAlpha(alpha)

        # Volume Modification
        elif action == VisualizationActions.VolumeModification:
            data = options['data']
            if isinstance(data, bool):
                itemObject.setFront2BackDrawing(data)
            elif len(data) == 2:
                itemObject.setThreshold(data[0])
                itemObject.setAlpha(data[1])
            else:
                raise TypeError('Unrecognise data')

        # Shader selection
        elif action == VisualizationActions.ShaderSelection:
            selected_shader = options['data'][0]

            if index == -1:
                for i in objectRoot:
                    i.setSelectedShader(selected_shader)
            else:
                itemObject.setSelectedShader(selected_shader)
        # Focus camera on object
        elif action == VisualizationActions.FocusObject:
            size = None
            center = itemObject.get_center()
            try:
                size = np.max(itemObject.get_size()) * 2
            except NotImplementedError:
                print(f"Object {itemObject} does not implement get_size.")
            finally:
                self.camera.focus(center, min_r=size)
                for shaders in self.shaderList:
                    for shader in shaders:
                        self._configView(shader)
        elif action == VisualizationActions.ColorSelection:
            data = options['data']

            if index == -1:
                for i in objectRoot:
                    i.set_color(data)
            else:
                itemObject.set_color(data)
        else:
            raise TypeError('Unrecognise visualization objects action')

        self.update()


    @QtCore.pyqtSlot(dict)
    def selectedObject(self, options):
        self.currentObject = options.get('current')

    @QtCore.pyqtSlot(dict)
    def modifySegmentation(self, options):
        # Makes the context the current one
        self.makeCurrent()

        segmentation = options.get('reference')
        attributes = options.get('attributes')
        segmentSubject = options.get('segmentSubject')
        updateViewingObjects = options.get('updateViewingObjects')
        updateSettingWindow = options.get('updateSettingWindow')

        if not isinstance(segmentation, SegmentationHandler):
            raise TypeError('modfySegmentation got an unxpected object reference: ', type(segmentation))

        if attributes:
            for attr in attributes.keys():
                try:
                    getattr(segmentation, attr)(*attributes[attr])
                except Exception as e:
                    QtWidgets.QMessageBox.warning(self, "Error", str(e), QtWidgets.QMessageBox.Ok)

        if segmentSubject:
            segmentation.segmentSubject()

        if updateViewingObjects:
            self.updateViewingObjects.emit()

        if updateSettingWindow:
            self.updateSettingWindow.emit(segmentation)

        self.update()


    def _configLight(self, shader, attributes):
        lightPos = attributes['lightPos']
        light = attributes['lightAttr']
        lightLa = light[0]
        lightLd = light[1]
        lightLs = light[2]
        material = attributes['materialAttr']
        materialKa = material[0]
        materialKd = material[1]
        materialKs = material[2]
        shininess = attributes['shininess']

        shader.glUseProgram()

        # Light
        GL.glUniform4f(shader.glGetUniformLocation("Light.pos"), *lightPos, 1.0)
        GL.glUniform3f(shader.glGetUniformLocation("Light.La"), lightLa, lightLa, lightLa)
        GL.glUniform3f(shader.glGetUniformLocation("Light.Ld"), lightLd, lightLd, lightLd)
        GL.glUniform3f(shader.glGetUniformLocation("Light.Ls"), lightLs, lightLs, lightLs)

        # Material
        GL.glUniform3f(shader.glGetUniformLocation("Material.Ka"), materialKa, materialKa, materialKa)
        GL.glUniform3f(shader.glGetUniformLocation("Material.Kd"), materialKd, materialKd, materialKd)
        GL.glUniform3f(shader.glGetUniformLocation("Material.Ks"), materialKs, materialKs, materialKs)
        GL.glUniform1f(shader.glGetUniformLocation("Material.shininess"), shininess)


    def _configPerspective(self, shader):
        shader.glUseProgram()
        GL.glUniformMatrix4fv(shader.glGetUniformLocation("P"),
                              1, GL.GL_TRUE, self.proj.getA())

    def _configView(self, shader, camera=None):
        if camera == None:
            camera = self.camera
        shader.glUseProgram()
        GL.glUniformMatrix4fv(shader.glGetUniformLocation("V"),
                              1, GL.GL_FALSE, camera.view.getA())



    def mousePressEvent(self, event):
        """ Captura el evento del boton del mouse clickeado y selecciona la funcion correspondiente
                a este. """
        self.screenX, self.screenY = event.pos().x(), event.pos().y()

        if event.button() == QtCore.Qt.LeftButton:
            self.orbit = True
            self.pan = False
            self.tilt = False
            event.accept()

        elif event.button() == QtCore.Qt.RightButton: # Permite desplazar el modelo.
            self.orbit = False
            self.pan = True
            self.panning = False
            self.tilt = False
            event.accept()

        elif event.button() == QtCore.Qt.MiddleButton:
            self.orbit = False
            self.pan = False
            self.tilt = True
            event.accept()

        else:
            self.orbit = False
            self.pan = False
            self.tilt = False
            event.ignore()


    def mouseMoveEvent(self, event):
        """ Ejecuta la funcion correspondiente al boton del mouse clickeado. """
        newScreenX, newScreenY = event.pos().x(), event.pos().y()
        dx, dy = newScreenX - self.screenX, newScreenY - self.screenY

        self.screenX, self.screenY = newScreenX, newScreenY

        if QtWidgets.QApplication.keyboardModifiers() & QtCore.Qt.ControlModifier:
            if not self.currentObject:
                return

            if self.orbit:
                axis, angle = self.camera.axisAndAngleFromScreen(dx, dy)

                if isinstance(self.currentObject, list):
                    for cObject_i in self.currentObject:
                        cObject_i.rotate(cObject_i.get_center(), angle, axis)
                else:
                    self.currentObject.rotate(self.currentObject.get_center(), angle, axis)

            elif self.pan:
                self.panning = True
                translateV = self.camera.vectorFromScreen(dx, -dy)

                if isinstance(self.currentObject, list):
                    for cObject_i in self.currentObject:
                        cObject_i.stackTranslate(translateV)
                else:
                    self.currentObject.stackTranslate(translateV)
            elif self.tilt:
                axis = self.camera.tiltAxisFromScreen()
                if isinstance(self.currentObject, list):
                    for cObject_i in self.currentObject:
                        cObject_i.spin(dx * np.pi / 10, axis)
                else:
                    self.currentObject.spin(dx *np.pi / 10, axis)
            self.objectsChanged.emit([self.currentObject])
        else:
            if self.orbit:
                if True:
                    self.camera.orbit(dx, dy)
                    self.csCamera.orbit(dx, dy)
                self._configView(self.coordSystemShader[0], self.csCamera)
            elif self.pan:
                self.panning = True
                self.camera.panning(dx, dy)
            elif self.tilt:
                self.camera.spin(np.pi * dx / 10)
                self.csCamera.spin(np.pi * dx / 10)
                self._configView(self.coordSystemShader[0], self.csCamera)
            else:
                return

            for s in self.shaderList:
                for shader in s:
                    self._configView(shader)

        event.accept()
        self.update()


    def wheelEvent(self, event):
        if QtWidgets.QApplication.keyboardModifiers() & QtCore.Qt.ControlModifier:
            if not self.currentObject:
                return

            delta = event.angleDelta().y()

            if isinstance(self.currentObject, list):
                for cObject_i in self.currentObject:
                    cObject_i.stackScale(delta*SCALE_FACTOR+1.0, self.camera.center)
            else:
                self.currentObject.stackScale(delta*SCALE_FACTOR+1.0, self.camera.center)

            self.objectsChanged.emit([self.currentObject])

        else:
            self.camera.zooming(event.angleDelta().y())

            for s in self.shaderList:
                for shader in s:
                    self._configView(shader)

        event.accept()
        self.update()

    def set_draw_bbox(self, value=None):
        if value is None:
            self.draw_boundingbox = not self.draw_boundingbox
        else:
            self.draw_boundingbox = value
        for bundle in self.bundles:
            bundle.setDrawBB(self.draw_boundingbox)
        for mri in self.mris:
            mri.setDrawBB(self.draw_boundingbox)
        for mesh in self.meshes:
            mesh.setDrawBB(self.draw_boundingbox)


    def keyPressEvent(self, event):
        key = event.key()

        if key == QtCore.Qt.Key_1:
            self.camera.frontView()
            self.csCamera.frontView()
        elif key == QtCore.Qt.Key_2:
            self.camera.backView()
            self.csCamera.backView()
        elif key == QtCore.Qt.Key_3:
            self.camera.leftView()
            self.csCamera.leftView()
        elif key == QtCore.Qt.Key_4:
            self.camera.rightView()
            self.csCamera.rightView()
        elif key == QtCore.Qt.Key_5:
            self.camera.topView()
            self.csCamera.topView()
        elif key == QtCore.Qt.Key_6:
            self.camera.bottomView()
            self.csCamera.bottomView()
        elif key == QtCore.Qt.Key_B:
            self.set_draw_bbox()
        elif key == QtCore.Qt.Key_W:
            self.camera.panning(0, 0, 10)
        elif key == QtCore.Qt.Key_A:
            self.camera.panning(10, 0)
        elif key == QtCore.Qt.Key_S:
            self.camera.panning(0, 0, -10)
        elif key == QtCore.Qt.Key_D:
            self.camera.panning(-10, 0)
        elif key == QtCore.Qt.Key_R:
            self.camera.panning(0, 10)
        elif key == QtCore.Qt.Key_F:
            self.camera.panning(0, -10)
        elif key == QtCore.Qt.Key_Q:
            self.camera.spin(np.pi * 1)
            self.csCamera.spin(np.pi * 1)
        elif key == QtCore.Qt.Key_E:
            self.camera.spin(np.pi * -1)
            self.csCamera.spin(np.pi * -1)
        elif key == QtCore.Qt.Key_Up:
            self.camera.orbit(0, 1)
            self.csCamera.orbit(0, 1)
        elif key == QtCore.Qt.Key_Down:
            self.camera.orbit(0, -1)
            self.csCamera.orbit(0, -1)
        elif key == QtCore.Qt.Key_Left:
            self.camera.orbit(1, 0)
            self.csCamera.orbit(1, 0)
        elif key == QtCore.Qt.Key_Right:
            self.camera.orbit(-1, 0)
            self.csCamera.orbit(-1, 0)
        elif key == QtCore.Qt.Key_F4:
            self.debug_breakpoint()
        else:
            event.ignore()
            return

        for s in self.shaderList:
            for shader in s:
                self._configView(shader)

        self._configView(self.coordSystemShader[0], self.csCamera)


        event.accept()
        self.update()


    @QtCore.pyqtSlot(QtCore.QPoint)
    def ShowContextMenu(self, pos):
        if self.panning:
            return

        self.makeCurrent()

        contextMenu = QtWidgets.QMenu()

        resetCameraAction = contextMenu.addAction("Reset camera")

        action = contextMenu.exec(self.mapToGlobal(pos))

        if action == resetCameraAction:
            self.camera.defaultValues()
            self.csCamera.defaultValues()

            for s in self.shaderList:
                for shader in s:
                    self._configView(shader)

            self._configView(self.coordSystemShader[0], self.csCamera)

        self.update()



    def testingModule(self):
        print('testingModule')
        self.segmentationTest()


    def cleanOpenGL(self):
        self.makeCurrent()

        for bundle in self.bundles:
            bundle.cleanOpenGL()

        for mesh in self.meshes:
            mesh.cleanOpenGL()

        for mri in self.mris:
            mri.cleanOpenGL()

        for roi in self.rois:
            roi.cleanOpenGL()

        self.coordSystem.cleanOpenGL()


    def performanceTestFPS(self):
        self.makeCurrent()
        testingN = 1000
        n = 100
        self.testing_i = 0
        self.testing = True
        self.testArray = np.empty(testingN*n, dtype=np.float32)
        for i in range(testingN):
            self.camera.orbit(10, 0)
            self.csCamera.orbit(10, 0)

            self._configView(self.coordSystemShader[0], self.csCamera)

            for s in self.shaderList:
                for shader in s:
                    self._configView(shader)

            for j in range(n):
                self.paintGL()

        self.testing = False

        print("Test ended. Number of samples taken: ", self.testing_i, "\nAverage frame time:", np.mean(self.testArray), "\nAverage fps: ", 1/np.mean(self.testArray))


    def segmentationTest(self):
        self.makeCurrent()
        try:
            testingN = 1000
            n = 10
            self.testing_i = 0
            self.testArray = np.empty(testingN*n, dtype=np.float32)

            for i in range(testingN):
                for j in range(n):
                    tb = time.time()
                    self.bundles[0].children[0].segmentSubject()
                    te = time.time()

                    self.testArray[self.testing_i] = te-tb
                    self.testing_i += 1

            print("Test ended. Number of samples taken: ", self.testing_i, "\nAverage time:", np.mean(self.testArray))
        except Exception as e:
            print(e)

    def debug_breakpoint(self):
        if not self.currentObject:
            return
        self.objectsChanged.emit([self.currentObject])

        for s in self.shaderList:
            for shader in s:
                self._configView(shader)



import nibabel as nib
import numpy as np
from pathlib import Path
from OpenGL import GL
from .VisualizationBaseObject import VisualizationBaseObject, propagateToChildren, drawable
from .Tools.visualizationEnums import VisualizationObject
from .Tools import glm
from .BoundingBox import BoundingBox

class MRI(VisualizationBaseObject):
    def __init__(self, sPath, shaderDict, parent):
        ''' Initialize a MRI object.
        It copies references to the path and the gl program.
        Then it loads the information from the file (points, normals, etc).

        It prepares the color texture, then loads it.

        Sends the information to the GPU buffers.

        Initialize  all parameters.

        The drawable variable is set to True.

        Parameters
        ----------
        sPath : str
            String containing path for the bundle file (.bundles)
        shaderProgram : GLInt
            Reference to the gl program containing all the shaders

        '''
        super().__init__(parent, None)
        self.identifier = VisualizationObject.MRI

        self.path = sPath
        self.fileName = Path(sPath).name

        self.mri = None
        self.readMRI()

        # Boundingbox edge color
        # self.color = np.array([1,1,1,1], dtype=np.float32)

        # Initialize of texture id and loading of self.colorTable into the GPU
        self.texture = None
        self._loadTexture()

        # Creates VBO then populates them with the vertexIdx data.
        self.vbo = None
        self._loadBuffers()

        # Ready to draw
        self.drawable = True

        # We set the boundingbox parameters
        dims = self.mriData.shape
        center = np.array([0, 0, 0], dtype=np.float32)

        self.boundingbox = BoundingBox(shaderDict, self, dims, center)

        print("Loading ready:\n\tDimensions of MRI: ", self.mriData.shape)


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteBuffers(1, [self.vbo])

        GL.glDeleteTextures([self.texture])

        self.clean = True

    def get_center(self):
        return self.boundingbox.get_center() + self.translateMat[:3, 3].flat

    def get_size(self):
        return self.boundingbox.get_size()

    def readMRI(self):
        '''

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        extension = Path(self.path).suffix

        # NIfTI file
        if extension == '.gz' or extension == '.nii':
            self._openNIfTI()

        # DICOM file
        # elif extension == 'dicom':
        #       self.openDICOM()

        # Unsopported file
        else:
            raise TypeError('Unsupported file.')

        self.resetModel()


    def _openNIfTI(self):
        '''

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        print('Opening MRI file {}'.format(self.path))

        self.mri = nib.load(self.path)
        self.mriData = np.array(self.mri.get_fdata(), dtype=np.float32).reshape(self.mri.get_fdata().shape)
        self.affine = np.asmatrix(self.mri.affine.reshape((4,4))).astype(np.float32)

        # print(self.mri.get_qform(coded=True))

        # # TESTING
        # self.mriData = np.swapaxes(self.mriData,0,1)
        # self.mriData = np.swapaxes(self.mriData,0,2)
        # # self.mriData = np.swapaxes(self.mriData,1,2)

        # self.mriData = self.mriData[::-1,:,:]
        # # self.mriData = self.mriData[:,::-1,:]
        # self.mriData = self.mriData[:,:,::-1]
        # self.affine = np.eye(4,dtype=np.float32)
        # self.affine[0][0] = 1.1
        # # self.affine = self.affine.T
        # img = nib.Nifti1Image(self.mriData, self.affine)
        # nib.save(img, self.fileName)

        print(self.affine)
        self.affine[0,3]=0
        self.affine[1,3]=0
        self.affine[2,3]=0
        # print(self.mri.affine)

        self.activeTransform = self.affine
        self.model = self.model*self.activeTransform


    def _loadTexture(self):
        if self.texture == None:
            self.texture = GL.glGenTextures(1)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_3D, self.texture)

        GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_R, GL.GL_CLAMP_TO_BORDER)

        bgColor = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float32)
        GL.glTexParameterfv(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_BORDER_COLOR, bgColor)

        # Filtered
        GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)

        # Not filtered
        # GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        # GL.glTexParameteri(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

        # Must swap axes 0 and 2, cuz data is not alined
        GL.glTexImage3D(GL.GL_TEXTURE_3D, 0, GL.GL_R32F, *self.mriData.shape[:3], 0, GL.GL_RED, GL.GL_FLOAT, np.swapaxes(self.mriData, 0, 2))


    def _loadBuffers(self):
        # Buffer with the texture data. Slice and Volume visualization objects
        self.vbo = GL.glGenBuffers(1)

        vertexIdx = np.arange(6, dtype=np.int32)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)
        GL.glBufferData(GL.GL_ARRAY_BUFFER, vertexIdx.nbytes, vertexIdx, GL.GL_STATIC_DRAW)


    def rotate(self, center, angle, axis):
        rotate = glm.translateMatrix(center)*glm.rotateMatrix(angle, axis)*glm.translateMatrix(-center)

        self.rotationMat = rotate*self.rotationMat

        self.model = self.translateMat*self.rotationMat*self.scaleMat*self.activeTransform
        self.inverseModel = np.linalg.inv(self.model)


    def translate(self, vec):
        self.translateMat = glm.translateMatrix(vec)

        self.model = self.translateMat*self.rotationMat*self.scaleMat*self.activeTransform
        self.inverseModel = np.linalg.inv(self.model)


    def stackTranslate(self, vec):
        newTranslateMat = glm.translateMatrix(vec)

        self.translateMat = newTranslateMat*self.translateMat

        self.model = self.translateMat*self.rotationMat*self.scaleMat*self.activeTransform
        self.inverseModel = np.linalg.inv(self.model)


    def scale(self, vec, center):
        self.scaleMat = glm.translateMatrix(center)*glm.scaleMatrix(vec)*glm.translateMatrix(-center)

        self.model = self.translateMat*self.rotationMat*self.scaleMat*self.activeTransform
        self.inverseModel = np.linalg.inv(self.model)


    def stackScale(self, vec, center):
        newScaleMat = glm.translateMatrix(center)*glm.scaleMatrix(vec)*glm.translateMatrix(-center)

        self.scaleMat = newScaleMat*self.scaleMat

        self.model = self.translateMat*self.rotationMat*self.scaleMat*self.activeTransform
        self.inverseModel = np.linalg.inv(self.model)


    def resetModel(self):
        self.scaleMat = glm.identity()

        self.translateMat = glm.identity()

        self.rotationMat = glm.identity()

        self.model = glm.identity()*self.activeTransform
        self.inverseModel = glm.identity()


    @propagateToChildren
    @drawable
    def draw(self):
        self.boundingbox.draw()

    @staticmethod
    def validExtension():
        return ['gz', 'nii']



import numpy as np

class Quaternion:
    ''' Class for storing rotations as quaternions
    '''

    def __init__(self):
        ''' Initializes a unit quaternion without rotation.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        self.w = 1
        self.x = 0
        self.y = 0
        self.z = 0

    @classmethod
    def fromAngleAxis(cls, angle, axis):
        ''' Initialices a quaternion from an angle and axis.

        The quaternion is normalized.

        Parameters
        ----------
        angle : float
            Angle of the rotation in radians

        axis : list, tuple, numpy.array
            A 3 element parameters with the axis of rotation. Not necessarilly normalized.

        Returns
        -------
        q : Quaternion
            Initialized quaternion

        '''

        module = np.sqrt(axis[0]**2 + axis[1]**2 + axis[2]**2)
        s = np.sin(angle/2) / module

        q = cls()

        q.w = np.cos(angle/2)
        q.x = float(axis[0]) * s
        q.y = float(axis[1]) * s
        q.z = float(axis[2]) * s

        return q

    @classmethod
    def fromQuaternionElements(cls, w, x, y, z):
        ''' Initializes a quaternion from the given elements.

        Parameters
        ----------
        w : float
            Real part of quaternion.

        x : float
            Imaginary part of quaternion, corresponding x coordinate.

        y : float
            Imaginary part of quaternion, corresponding y coordinate.

        z : float
            Imaginary part of quaternion, corresponding z coordinate.

        Returns
        -------
        q : Quaternion
            Initialized quaternion

        '''

        q = cls()

        q.w = float(w)
        q.x = float(x)
        q.y = float(y)
        q.z = float(z)

        return q

    @classmethod
    def fromQuaternion(cls, other):
        ''' Initializes a quaternion from another quaternion.

        Parameters
        ----------
        other : Quaternion
            Quaternion to be copied.

        Returns
        -------
        q : Quaternion
            Initialized quaternion

        '''

        q = cls()

        q.w = other.w
        q.x = other.x
        q.y = other.y
        q.z = other.z

        return q

    @classmethod
    def identity(cls):
        ''' Initializes a unit quaternion without rotation.

        Parameters
        ----------
        None

        Returns
        -------
        q : Quaternion
            Initialized quaternion

        '''

        q = cls()

        q.w = 1.0
        q.x = 0.0
        q.y = 0.0
        q.z = 0.0

        return q

    @classmethod
    def from_pair_of_vectors(cls, vector_1, vector_2):
        ''' Initializes a quaternion from a pair of vectors.

        Parameters
        ----------
        vector_1 : list, tuple, numpy.ndarray
            First vector of 3 elements.
        vector_2 : list, tuple, numpy.ndarray
            Second vector of 3 elements.

        Returns
        -------
        q : Quaternion
            Initialized quaternion
        '''
        c_prod = np.cross(vector_1, vector_2)
        n1, n2, nc_prod = np.linalg.norm([vector_1, vector_2, c_prod], axis=1)
        angle = np.arcsin(nc_prod / n1 / n2)
        axis = np.arcsin(c_prod / n1 / n2)
        return cls.fromAngleAxis(angle, axis)


    def normalize(self):
        ''' Normalizes the quaternion.

        Parameters
        ----------
        None

        Returns
        -------
        self : Quaternion

        '''

        module = np.sqrt(self.w**2 + self.x**2 + self.y**2 + self.z**2)
        self.w /= module
        self.x /= module
        self.y /= module
        self.z /= module

        return self


    def inverted(self):
        ''' Invertes the quaternion, by inverting the axis.

        Parameters
        ----------
        None

        Returns
        -------
        q : Quaternion
            Inverted quaternion.

        '''

        return Quaternion.fromQuaternionElements(self.w, -self.x, -self.y, -self.z)


    def rotateVector(self, vector):
        ''' Rotates a vector using the quaternion.

        Parameters
        ----------
        vector : list, tuple, numpy.array
            A 3 element parameters with the axis of rotation. Not necessarilly normalized.

        Returns
        -------
        array : numpy.array
            Rotated vector.

        '''

        p = Quaternion.fromQuaternionElements(0, vector[0], vector[1], vector[2])
        pp = self * p * self.inverted()

        return np.array([pp.x, pp.y, pp.z], dtype=np.float32)


    def rotation3Matrix(self):
        ''' Creates a 3x3 rotation matrix from the quaternion.

        Parameters
        ----------
        None

        Returns
        -------
        matrix : numpy.matrix
            Rotation matrix.

        '''

        xx = self.x**2
        yy = self.y**2
        zz = self.z**2
        xy = self.x*self.y
        wz = self.w*self.z
        xz = self.x*self.z
        wy = self.w*self.y
        yz = self.y*self.z
        wx = self.w*self.x

        return np.matrix([	[1-2*(yy+zz),	2*(xy-wz),		2*(xz+wy)],
                            [2*(xy+wz),		1-2*(xx+zz),	2*(yz-wx)],
                            [2*(xz-wy),		2*(yz+wx),		1-2*(xx+yy)]], dtype=np.float32)


    def rotation4Matrix(self):
        ''' Creates a 4x4 rotation matrix from the quaternion.

        Parameters
        ----------
        None

        Returns
        -------
        matrix : numpy.matrix
            Rotation matrix.

        '''

        xx = self.x**2
        yy = self.y**2
        zz = self.z**2
        xy = self.x*self.y
        wz = self.w*self.z
        xz = self.x*self.z
        wy = self.w*self.y
        yz = self.y*self.z
        wx = self.w*self.x

        return np.matrix([	[1-2*(yy+zz),	2*(xy-wz),		2*(xz+wy),		0],
                            [2*(xy+wz),		1-2*(xx+zz),	2*(yz-wx),		0],
                            [2*(xz-wy),		2*(yz+wx),		1-2*(xx+yy),	0],
                            [0,				0,				0,				1]], dtype=np.float32)


    def toAngleAxis(self):
        ''' Transforms the quaternion into the angle and axis of rotation in 3D coordinate system.

        Parameters
        ----------
        None

        Returns
        -------
        angle : float
            Angle of rotation.

        axis : float
            axis of rotation.

        '''

        angle = np.arccos(self.w)*2
        axis = np.array([self.x, self.y, self.z], dtype=np.float32) / np.sqrt(1-self.w**2)

        return angle, axis


    def slerp(self, other, t):
        ''' Spherical linear interpolation from the given quaternion to another, by the given t portion.

        Parameters
        ----------
        other : Quaternion
            Final quaternion to which the interpolation goes.

        t : float
            Step from the linear interpolation between both quaternions. 0 <= t <= 1.

        Returns
        -------
        q : Quaternion
            New quaternion with the slerp rotation in t.

        '''

        return (other*self.inverted())**t * self


    def __mul__(self, other):
        ''' Overload multiply, for multiplying quaternions between them.

        Parameters
        ----------
        other : Quaternion
            Rotation to be apply.

        Returns
        -------
        q : Quaternion
            New quaternion with the rotation applied.

        '''

        if not isinstance(other, Quaternion):
            raise TypeError('Can not multiply with other object that is not a Quaternion')

        newW = self.w*other.w - self.x*other.x - self.y*other.y - self.z*other.z
        newX = self.x*other.w + self.w*other.x - self.z*other.y + self.y*other.z
        newY = self.y*other.w + self.z*other.x + self.w*other.y - self.x*other.z
        newZ = self.z*other.w - self.y*other.x + self.x*other.y + self.w*other.z

        return Quaternion.fromQuaternionElements(newW, newX, newY, newZ)


    def __rmul__(self, other):
        ''' Bypass to normal multiply.

        Parameters
        ----------
        other : Quaternion
            Rotation to be apply.

        Returns
        -------
        q : Quaternion
            New quaternion with the rotation applied.

        '''

        return self*other
        # raise TypeError('Can not multiply with other object that is not a Quaternion')


    def __pow__(self, t):
        ''' Scales the angle of rotation in t.

        Parameters
        ----------
        t : float
            Floating point to scale the angle of rotation of the quaternion.

        Returns
        -------
        q : Quaternion
            New quaternion with the same axis, but angle scaled in t.

        '''

        angle, axis = self.toAngleAxis()

        return Quaternion.fromAngleAxis(angle*t, axis)


    def __str__(self):
        ''' String representation of the quaternion.

        Parameters
        ----------
        None

        Returns
        -------
        string : str
            w, x, y, z as string representation of list.

        '''

        return '[{0}, {1}, {2}, {3}]'.format(self.w, self.x, self.y, self.z)

    def __iter__(self):
        ''' Iterable representation of the quaternion.

        Parameters
        ----------
        None

        Returns
        -------
        iter : iterable
            yields w, x, y and z.

        '''

        yield self.w
        yield self.x
        yield self.y
        yield self.z


'''
'''
import random
import numpy as np
import ctypes as ct
from OpenGL import GL
from ..Tools import glm
from ..VisualizationBaseObject import VisualizationBaseObject, VisualizationObject, drawable, config
from ..Shaders import Shader
from ..Tools.visualizationEnums import SegmentationTypes, ROIType
from ..Tools.Quaternion import Quaternion
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_b_vs = files('phybers.fibervis') / 'shaders' / 'bundle.vs'
_sfs_vs = files('phybers.fibervis') / 'shaders' / 'standardFragmentShader.fs'

class ROI(VisualizationBaseObject):
    '''
    '''

    def __init__(self, roitype, center, radius, shaderDict, parent):
        '''
        '''

        if not roitype in ROIType:
            raise TypeError('Unsupported ROI type.')

        super().__init__(parent, shaderDict)
        self.identifier = VisualizationObject.ROI

        self.fileName = str(roitype).split('.')[-1]

        self.roiType = roitype

        self.center = np.array(center, dtype=np.float32)

        self.radius = np.array(radius, dtype=np.float32)

        self.squaredRadius  = self.radius**2

        self.points = None
        self.normals = None
        self.color = None
        self.createROI()

        # Initialize the values on colorTable
        self.colorTable = np.array([random.random(), random.random(), random.random(), 1.0], dtype=np.float32)

        # Initialize of texture id and loading of self.colorTable into the GPU
        self.colorTableTexture = None
        self._loadColorTexture()

        # Creates VBOs & an EBO, then populates them with the point, normal and color data. The elements data goes into the EBO
        self._loadBuffers()

        # Random translation
        self.translate([random.random()*100 for x in range(3)])

        # Ready to draw
        self.drawable = True#parent.drawable

    def get_center(self):
        return self.center

    def get_size(self):
        return (self.radius * 2).repeat(3)


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(1, [self.vao])

        GL.glDeleteBuffers(1, [self.vbo])

        GL.glDeleteTextures([self.colorTableTexture])

        self.clean = True


    def createROI(self):
        if self.roiType == ROIType.Sphere:
            self.createSphere()
            self.translation = glm.translateMatrix(self.center)
            self.scalation = glm.scaleMatrix((self.radius, self.radius, self.radius))
        elif self.roiType == ROIType.Aabb:
            raise NotImplementedError('ROI type not implemented: ', self.roiType)
        elif self.roiType == ROIType.Obb:
            raise NotImplementedError('ROI type not implemented: ', self.roiType)
        elif self.roiType == ROIType.Plane:
            raise NotImplementedError('ROI type not implemented: ', self.roiType)


    def _loadColorTexture(self):
        ''' It generates a texture (if not already). Then makes the texture0 the active one and loads the color table as a 1D texture.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        self.validBundleColorTexDims = (np.int32(1), np.int32(1))

        if self.colorTableTexture == None:
            self.colorTableTexture = GL.glGenTextures(1)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)

        bgColor = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float32)
        GL.glTexParameterfv(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_BORDER_COLOR, bgColor)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

        # GL.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_RGBA, 1, 0, GL.GL_RGBA, GL.GL_FLOAT, self.colorTable.flatten())
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, *self.validBundleColorTexDims, 0, GL.GL_RGBA, GL.GL_FLOAT, self.colorTable.flatten())


    def _loadBuffers(self):
        ''' It generates a VAO, VBO and EBO.
        It populates them with the vertex, normal and colors (the VBO), also the elements (the EBO).

        It finishes when the buffers are linked with attributes on the vertex shader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        # Reference for vbos, ebos and attribute ptrs
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        self.vbo = GL.glGenBuffers(1)

        # VBO
        bufferSize = self.points.nbytes + self.normals.nbytes + self.color.nbytes
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)
        GL.glBufferData(GL.GL_ARRAY_BUFFER, bufferSize, None, GL.GL_STATIC_DRAW)        # Create empty buffer

        # Populate buffer with points
        GL.glBufferSubData(GL.GL_ARRAY_BUFFER,
            0,
            self.points.nbytes,
            self.points.flatten())

        # Normals
        GL.glBufferSubData(GL.GL_ARRAY_BUFFER,
            self.points.nbytes,
            self.normals.nbytes,
            self.normals.flatten())

        # Color-bundle id
        GL.glBufferSubData(GL.GL_ARRAY_BUFFER,
            (2*self.points.nbytes),
            self.color.nbytes,
            self.color.flatten())

        # Enable attributes
        positionAttribute =     self.shader[0].attributeLocation('vertexPos')
        normalAttribute =       self.shader[0].attributeLocation('vertexNor')
        colorAttribute =        self.shader[0].attributeLocation('vertexCol')

        # # Connect attributes
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute,3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        GL.glEnableVertexAttribArray(normalAttribute)
        GL.glVertexAttribPointer(normalAttribute,       3, GL.GL_FLOAT, GL.GL_FALSE, 0, ct.c_void_p(self.points.nbytes))

        GL.glEnableVertexAttribArray(colorAttribute)
        GL.glVertexAttribPointer(colorAttribute,        1, GL.GL_INT, GL.GL_FALSE, 0, ct.c_void_p(2*self.points.nbytes))

    def createSphere(self):
        def subdivideTriangles(triangles):
            result = []

            for triangle in triangles:
                # first triangle
                result.append([ triangle[0],
                                (triangle[1]-triangle[0])*0.5 + triangle[0],
                                (triangle[2]-triangle[0])*0.5 + triangle[0]])

                # second triangle
                result.append([ triangle[1],
                                (triangle[0]-triangle[1])*0.5 + triangle[1],
                                (triangle[2]-triangle[1])*0.5 + triangle[1]])

                # third triangle
                result.append([ triangle[2],
                                (triangle[0]-triangle[2])*0.5 + triangle[2],
                                (triangle[1]-triangle[2])*0.5 + triangle[2]])

                # fourth triangle
                result.append([ (triangle[1]-triangle[0])*0.5 + triangle[0],
                                (triangle[2]-triangle[0])*0.5 + triangle[0],
                                (triangle[2]-triangle[1])*0.5 + triangle[1]])

            return np.array(result, dtype=np.float32)

        triangles = np.array([[[1.0, 0.0, 0.0],
                               [0.0, 1.0, 0.0],
                               [0.0, 0.0, 1.0]]], dtype=np.float32)

        # We subdivide the triangle in 4 triangles
        for i in range(3):
            triangles = subdivideTriangles(triangles)

        # We normalize the sphere vertex, so we have also the normals
        for triangle in triangles:
            for vertex in triangle:
                vertex /= np.sqrt((vertex**2).sum())

        # We create the half sphere, by rotating the current 1/8
        rotated = np.copy(triangles)
        rotation = Quaternion.fromAngleAxis(np.pi/2, [0.0,1.0,0.0])
        for i in range(3):
            for triangle in rotated:
                for vertex in triangle:
                    vertex[:] = rotation.rotateVector(vertex)
            triangles = np.concatenate((triangles,rotated))

        # We create the whole sphere
        rotated = np.copy(triangles)
        rotated[:,:,1] *= -1
        triangles = np.concatenate((triangles,rotated))

        self.points = np.ravel(triangles)
        self.normals = self.points
        self.color = np.zeros(self.points.size//3, dtype=np.int32)


    def setContainsMethods(self):
        if self.roiType == ROIType.Sphere:
            self.containsVertex = self.containsVertexSphere
            self.containsBoundary = self.containsBoundarySphere
            self.intersectsBoundary = self.intersectsBoundarySphere

        elif self.roiType == ROIType.Aabb:
            raise TypeError('Contains method for ROI type not implemented: ', self.roiType)

        elif self.roiType == ROIType.Obb:
            raise TypeError('Contains method for ROI type not implemented: ', self.roiType)

        elif self.roiType == ROIType.Plane:
            raise TypeError('Contains method for ROI type not implemented: ', self.roiType)

    def containsBoundary(self, center, radius):
        print('method not being called')
        pass


    def containsVertex(self, vertex):
        print('method not being called')
        pass


    def intersectsBoundary(self, center, radius):
        print('method not being called')
        pass


    def containsBoundarySphere(self, center, radius):
        points = center + np.array([[ radius[0],  radius[1],  radius[2]],
                                    [-radius[0],  radius[1],  radius[2]],
                                    [ radius[0], -radius[1],  radius[2]],
                                    [-radius[0], -radius[1],  radius[2]],
                                    [ radius[0],  radius[1], -radius[2]],
                                    [-radius[0],  radius[1], -radius[2]],
                                    [ radius[0], -radius[1], -radius[2]],
                                    [-radius[0], -radius[1], -radius[2]]],dtype=np.float32)

        dist = points - self.center

        return ((dist**2).sum() <= self.squaredRadius).sum()


    def containsVertexSphere(self, vertex):
        dist = vertex - self.center

        return (dist**2).sum() <= self.squaredRadius


    def intersectsBoundarySphere(self, center, radius):
        squaredDist = 0.0

        if self.center[0] < center[0]-radius[0]:
            squaredDist += (self.center[0]-(center[0]-radius[0]))**2
        elif self.center[0] > center[0]+radius[0]:
            squaredDist += (self.center[0]-(center[0]+radius[0]))**2

        if self.center[1] < center[1]-radius[1]:
            squaredDist += (self.center[1]-(center[1]-radius[1]))**2
        elif self.center[1] > center[1]+radius[1]:
            squaredDist += (self.center[1]-(center[1]+radius[1]))**2

        if self.center[2] < center[2]-radius[2]:
            squaredDist += (self.center[2]-(center[2]-radius[2]))**2
        elif self.center[2] > center[2]+radius[2]:
            squaredDist += (self.center[2]-(center[2]+radius[2]))**2

        return squaredDist <= self.squaredRadius


    def loadUniforms(self):
        '''  This function is called always before drawing. It makes sure that the current bundle will be visualized with the uniforms specified for it.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        finalModel = self.model*self.translation*self.scalation
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("M"), 1, GL.GL_TRUE, finalModel.getA())
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation("colorTable"), 0)

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('texture1DMax'), self.validBundleColorTexDims[0])


    def getRGB256(self):
        return self.colorTable*255


    def getCenter(self, inverseModel = glm.identity()):
        finalModel = self.model*self.translation*self.scalation
        return (inverseModel * finalModel * np.array([0, 0, 0, 1], dtype=np.float32).reshape((4,1)))[:3].ravel()


    def getRadius(self, inverseModel = glm.identity()):
        finalModel = inverseModel * self.model*self.translation*self.scalation

        r = (finalModel * np.array([1, 0, 0, 1], dtype=np.float32).reshape((4,1)))[:3].ravel()
        r = r - self.getCenter(inverseModel)
        r = np.linalg.norm(r)

        return np.array([r, r, r], dtype=np.float32)


    def changeFileName(self, newFilename):
        self.fileName = newFilename


    def getROIValue(self):
        return self.roiType.value


    @drawable
    @config
    def draw(self):
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        GL.glDrawArrays(GL.GL_TRIANGLES, 0, self.color.size);

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_b_vs)]
        fragmentGLSL = [str(_sfs_vs)]
        return [Shader(vertexGLSL, fragmentGLSL)]



from PyQt5 import QtGui, QtWidgets, uic, QtCore
from functools import wraps
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
from ...Framework.VisualizationBaseObject import VisualizationObject

_roisd_ui = files('phybers.fibervis.ui') / 'Segmentations' / 'ROIsSegmentationDialog.ui'


# MUST DELETE
def identifyNumberRecursively(item, number):
    root = item

    if item.parent() != None:
        root = identifyNumberRecursively(item.parent(), number)
        number.append(item.parent().indexOfChild(item))
    else:
        return item
    return root


def retrieveItemIteratively(root, index):
    item = root

    for i in index:
        if hasattr(item, 'children'):
            item = item.children[i]
        else:
            item = item[i]

    return item
    ###########


def detectAlways(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)

        if args[0].ui.alwaysBox.isChecked():
            args[0].segmentSubject()
    return wrapper


class ROIsSegmentationDialog(QtWidgets.QDialog):
    updateObject = QtCore.pyqtSignal(dict)


    def __init__(self, segmentation, parent):
        super().__init__(parent)
        self.ui = uic.loadUi(str(_roisd_ui), self)

        self.segmentationRef = segmentation

        self.ui.tableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.ui.tableWidget.horizontalHeader().setStretchLastSection(True)

        self.ui.addROIButton.clicked.connect(self.addROI)
        self.ui.removeROIButton.clicked.connect(self.removeROI)

        self.ui.detectButton.clicked.connect(self.segmentSubject)

        self.ui.exportButton.clicked.connect(self.exportBundle)

        self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

        self.ui.tableWidget.cellChanged.connect(self.cellChanged)

        self.ui.logicLineEdit.editingFinished.connect(self.logicLineEditEdited)

        self.show()




    # // Actualiza dibujo al clickear checkbox de la tabla
    # this.connect(this.ui.tableWidget, SIGNAL(cellChanged(int,int)), this, SLOT(itemtoggle(int,int)));


    @detectAlways
    @QtCore.pyqtSlot()
    def addROI(self):
        treeWidgetItems = self.parentWidget().ui.treeWidget.selectedItems()
        if len(treeWidgetItems) == 0:
            return
        elif treeWidgetItems[0].identifier != VisualizationObject.ROI:
            QtWidgets.QMessageBox.question(self, "Invalid ROI", 'Must select a ROI from the visualization tool dialog.', QtWidgets.QMessageBox.Ok)
            return

        # Bad practice
        ##############
        number = []
        root = identifyNumberRecursively(treeWidgetItems[0], number)

        if root.text(0) == 'Bundles':
            reference = self.parentWidget().bundleReference
        elif root.text(0) == 'Meshes':
            reference = self.parentWidget().meshReference
        elif root.text(0) == 'MRIs':
            reference = self.parentWidget().mriReference
        elif root.text(0) == 'ROIs':
            reference = self.parentWidget().roisReference
        else:
            reference = None

        refObj = retrieveItemIteratively(reference, number)
        if isinstance(refObj, list):
            QtWidgets.QMessageBox.question(self, "Too many ROIs", 'Please select one ROI.', QtWidgets.QMessageBox.Ok)
            return
        ##############

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'addROI' : [refObj]},
                                'updateViewingObjects' : True}

        self.updateObject.emit(options)


    @detectAlways
    @QtCore.pyqtSlot()
    def removeROI(self):
        index = self.ui.tableWidget.currentRow()

        # self.ui.tableWidget.removeRow(index)
        options = {     'reference' : self.segmentationRef,
                                'attributes': {'removeROIFromIndex' : [index]},
                                'updateViewingObjects' : True}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def segmentSubject(self):
        options = {     'reference' : self.segmentationRef,
                                'segmentSubject' : True}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def alphaChanged(self):
        alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'setAlpha' : [alpha]}}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def exportBundle(self):

        # Gets direction of bundle files
        outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

        options = {     'reference' : self.segmentationRef,
                                'attributes' : {'exportFile' : [outFile]}}

        self.updateObject.emit(options)


    @detectAlways
    @QtCore.pyqtSlot(int, int)
    def cellChanged(self, row, column):
        if column == 0:
            checked = self.ui.tableWidget.item(row, column).checkState() == QtCore.Qt.Checked

            options = {     'reference' : self.segmentationRef,
                                    'attributes' : {'setValidatorAtIndex' : [row, checked]}}

            self.updateObject.emit(options)


    @detectAlways
    @QtCore.pyqtSlot()
    def logicLineEditEdited(self):
        options = {     'reference' : self.segmentationRef,
                                'attributes': {'updateLogic' : [str(self.ui.logicLineEdit.text()),
                                                                                        self.ui.logicLineEdit.setText]}}

        self.updateObject.emit(options)


    def update(self):
        while self.ui.tableWidget.rowCount():
            self.ui.tableWidget.removeRow(0)

        for i in range(len(self.segmentationRef.children)):
            self.insertItemInTable(self.segmentationRef.children[i], self.segmentationRef.roiValidator[i])

    def objectsChanged(self, objects):
        if self.ui.alwaysBox.isChecked():
            for obj in objects:
                if obj in self.segmentationRef.children or obj is self.segmentationRef:
                    self.segmentSubject()
                    return


    def insertItemInTable(self, objRef, checked):
        self.ui.tableWidget.cellChanged.disconnect(self.cellChanged)

        row = self.ui.tableWidget.rowCount()
        self.ui.tableWidget.insertRow(row)

        tag = QtWidgets.QTableWidgetItem(str(row))
        tag.setCheckState(QtCore.Qt.Checked if checked else QtCore.Qt.Unchecked)
        tag.setTextAlignment(QtCore.Qt.AlignHCenter)
        tag.setFlags(tag.flags() & ~QtCore.Qt.ItemIsEditable)

        bg = QtWidgets.QTableWidgetItem()
        bg.setBackground(QtGui.QBrush(QtGui.QColor(
                                        int(objRef.getRGB256()[0]),
                                        int(objRef.getRGB256()[1]),
                                        int(objRef.getRGB256()[2]))))
        bg.setFlags(bg.flags() & ~QtCore.Qt.ItemIsEditable)

        roiType = QtWidgets.QTableWidgetItem(str(objRef.roiType))
        roiType.setTextAlignment(QtCore.Qt.AlignHCenter)
        roiType.setFlags(roiType.flags() & ~QtCore.Qt.ItemIsEditable)

        self.ui.tableWidget.setItem(row, 0, tag)
        self.ui.tableWidget.setItem(row, 1, bg)
        self.ui.tableWidget.setItem(row, 2, roiType)

        self.ui.tableWidget.selectRow(row)

        self.ui.tableWidget.cellChanged.connect(self.cellChanged)



import numpy as np
from OpenGL import GL
from .VisualizationBaseObject import VisualizationObject, VisualizationBaseObject, drawable, config
from .Shaders import Shader
from .BoundingBox import BoundingBox
from skimage.filters import threshold_otsu
from .Tools import glm
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_vs_vs = files('phybers.fibervis') / 'shaders' / 'volume-slice.vs'
_vs_fs = files('phybers.fibervis') / 'shaders' / 'volume.fs'

class Volume(VisualizationBaseObject):
    def __init__(self, mri, shaderDict, camera):

        super().__init__(mri, shaderDict)
        self.identifier = VisualizationObject.MRIVolume

        self.fileName = 'Volume render'

        self.scaleModel = mri.affine * glm.scaleMatrix(mri.mriData.shape)

        # We select an otsu threshold for volume render and slope for grey color
        self.threshold = threshold_otsu(mri.mriData)
        self.slope = 1.0/mri.mriData.max()

        # Matrixs
        self.model = mri.translateMat*mri.rotationMat*mri.scaleMat
        self.inverseModel = np.linalg.inv(self.model)

        self.scaleMat = mri.scaleMat
        self.translateMat = mri.translateMat
        self.rotationMat = mri.rotationMat

        self.axis = None
        self.calculateAxisMat3()

        # Sampling frequency, twice the maximum frequency
        self.sliceFr = int(2*np.linalg.norm(mri.mriData.shape))

        self.loadStaticUniforms()

        # MRI data already loaded in the gpu
        self.texture = mri.texture
        self.vao = None
        self.vbo = mri.vbo

        self._loadBuffers()

        self.alpha = 0.5
        self.f2bDrawing = True

        self.np = None
        self.__camera = camera # for memory cleaning purpose only
        self.updateCamera(camera.eye)
        camera.setObject2Notify(self)

        # Ready to draw
        self.drawable = True

        self.boundingbox = BoundingBox(shaderDict, self, mri.boundingbox.dims, mri.boundingbox.center, self.scaleModel)

        self.max = mri.mriData.max()


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(1, [self.vao])

        self.clean = True


    def get_center(self):
        return self.boundingbox.get_center() + self.translateMat[:3, 3].flat

    def get_size(self):
        return self.boundingbox.get_size()

    def loadStaticUniforms(self):
        self.shader[0].glUseProgram()

        v1 = np.array([0,1,3,-1, 1,0,1,3, 0,4,5,-1, 4,0,4,5, 0,2,6,-1, 2,0,2,6], dtype=np.int32)
        v2 = np.array([1,3,7,-1, 5,1,3,7, 4,5,7,-1, 6,4,5,7, 2,6,7,-1, 3,2,6,7], dtype=np.int32)

        nSequence = np.array([	0,1,2,3,4,5,6,7, 1,3,0,2,5,7,4,6, 2,0,3,1,6,4,7,5, 3,2,1,0,7,6,5,4,
                                4,0,6,2,5,1,7,3, 5,1,4,0,7,3,6,2, 6,2,7,3,4,0,5,1, 7,6,3,2,5,4,1,0], dtype=np.int32)

        self.vertexPoints = np.array([	1,1,0,	1,1,1,	0,1,0,	0,1,1,
                                        1,0,0,	1,0,1,	0,0,0,	0,0,1], dtype=np.float32).reshape((8,3))

        self.uniDPlane = self.shader[0].glGetUniformLocation("dPlane")

        GL.glUniform1iv(self.shader[0].glGetUniformLocation("v1"), 24, v1)
        GL.glUniform1iv(self.shader[0].glGetUniformLocation("v2"), 24, v2)
        GL.glUniform1iv(self.shader[0].glGetUniformLocation("nSequence"), 64, nSequence)
        GL.glUniform3fv(self.shader[0].glGetUniformLocation("vertexPoints"), 8, self.vertexPoints)


    def _loadBuffers(self):
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)

        # Enable attributes
        vertexIdxAttribute = self.shader[self.selectedShader].attributeLocation('vertexIdx')

        # # Connect attributes
        GL.glEnableVertexAttribArray(vertexIdxAttribute)
        GL.glVertexAttribIPointer(vertexIdxAttribute, 1, GL.GL_INT, 0, None)


    def updateCamera(self, newEye):
        self.eye = newEye
        volumeCenter = (self.model*self.scaleModel * np.array([0.5, 0.5, 0.5, 1.0], dtype=np.float32).reshape((4,1)))[:3].getA().ravel()

        self.np = self.eye - volumeCenter
        self.np /= np.linalg.norm(self.np)

        self.dPlaneBegin = np.dot((self.model*self.scaleModel * np.array([*self.vertexPoints[0], 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel(), self.np)
        self.dPlaneEnd   = self.dPlaneBegin
        self.frontIdx = 0

        for i in range(1,8):
            D = np.dot((self.model*self.scaleModel * np.array([*self.vertexPoints[i], 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel(), self.np)

            if D > self.dPlaneBegin:
                self.dPlaneBegin = D
                self.frontIdx = i
            if D < self.dPlaneEnd:
                self.dPlaneEnd = D

        if self.f2bDrawing:
            self.dPlane = self.dPlaneBegin
            self.dPlaneStep = (self.dPlaneEnd - self.dPlaneBegin)/self.sliceFr
        else:
            self.dPlane = self.dPlaneEnd
            self.dPlaneStep = (self.dPlaneBegin - self.dPlaneEnd)/self.sliceFr


    def loadUniforms(self):
        '''  This function is called always before drawing. It makes sure that the current bundle will be visualized with the uniforms specified for it.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''
        # Vertex shader
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("M"), 1, GL.GL_TRUE, self.model.getA())
        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation("S"), 1, GL.GL_TRUE, self.scaleModel.getA())
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation("frontIdx"), self.frontIdx)
        GL.glUniform4f(self.shader[self.selectedShader].glGetUniformLocation("np"), *self.np, 0.0)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("dPlane"), self.dPlane)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("dPlaneStep"), self.dPlaneStep)

        # Fragment shader
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("slope"), self.slope)
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation("mriTexture"), 0)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation("threshold"), self.threshold)
        GL.glUniform3fv(self.shader[self.selectedShader].glGetUniformLocation('eye'), 1, self.eye)
        GL.glUniformMatrix3fv(self.shader[self.selectedShader].glGetUniformLocation('axis'), 1, GL.GL_TRUE, self.axis)
        GL.glUniform1f(self.shader[self.selectedShader].glGetUniformLocation('alpha'), self.alpha)


    def calculateAxisMat3(self):
        x = np.array([1,0,0,1], dtype=np.float32)
        y = np.array([0,1,0,1], dtype=np.float32)
        z = np.array([0,0,1,1], dtype=np.float32)

        begin = (self.model*self.scaleModel * np.array([0.0, 0.0, 0.0, 1.0], dtype=np.float32).reshape((4,1)))[:3].ravel()
        x = (self.model*self.scaleModel * x.reshape((4,1)))[:3].ravel()
        y = (self.model*self.scaleModel * y.reshape((4,1)))[:3].ravel()
        z = (self.model*self.scaleModel * z.reshape((4,1)))[:3].ravel()

        x -= begin
        y -= begin
        z -= begin

        x = (x / np.linalg.norm(x)).getA().ravel()[:3]
        y = -(y / np.linalg.norm(y)).getA().ravel()[:3]
        z = (z / np.linalg.norm(z)).getA().ravel()[:3]

        self.axis = np.array([*x, *y, *z], dtype=np.float32).reshape((3,3)).T


    # @propagateToChildren
    def clearRef(self):
        super().clearRef()
        self.__camera.removeObject2Notify(self)


    def setFront2BackDrawing(self, newf2bDrawing):
        self.f2bDrawing = newf2bDrawing

        if newf2bDrawing is True:
            self.dPlane = self.dPlaneBegin
            self.dPlaneStep = (self.dPlaneEnd - self.dPlaneBegin)/self.sliceFr
        else:
            self.dPlane = self.dPlaneEnd
            self.dPlaneStep = (self.dPlaneBegin - self.dPlaneEnd)/self.sliceFr


    def setAlpha(self, newAlpha):
        self.alpha = newAlpha


    def setThreshold(self, newThreshold):
        self.threshold = newThreshold


    @drawable
    @config
    def draw(self):
        GL.glEnable(GL.GL_BLEND)
        GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_3D, self.texture)

        GL.glDrawArraysInstanced(GL.GL_TRIANGLE_FAN, 0, 6, self.sliceFr)
        GL.glDisable(GL.GL_BLEND)

        self.boundingbox.draw()

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_vs_vs)]
        fragmentGLSL = [str(_vs_fs)]
        return [Shader(vertexGLSL, fragmentGLSL)]


    def rotate(self, center, angle, axis):
        super().rotate(center, angle, axis)
        self.updateCamera(self.eye)
        self.calculateAxisMat3()


    def translate(self, vec):
        super().translate(vec)
        self.updateCamera(self.eye)
        self.calculateAxisMat3()


    def stackTranslate(self, vec):
        super().stackTranslate(vec)
        self.updateCamera(self.eye)
        self.calculateAxisMat3()


    def scale(self, vec, center):
        super().scale(vec)
        self.updateCamera(self.eye)
        self.calculateAxisMat3()


    def stackScale(self, vec, center):
        super().stackScale(vec, center)
        self.updateCamera(self.eye)
        self.calculateAxisMat3()


    def resetModel(self):
        super().resetModel()
        self.updateCamera(self.eye)
        self.calculateAxisMat3()


import random
import math
import numpy as np
import networkx as nx
import multiprocessing as mp
from OpenGL import GL
from sklearn.cluster import MiniBatchKMeans
from scipy import sparse
from functools import partial
from .SegmentationHandler import SegmentationHandler
from ..Tools.utilities import findIntegersMultiplierFor
from ..Tools.visualizationEnums import SegmentationTypes
from ..BoundingBox import BoundingBox
from ...FiberVis_core import ffclust, reSampleBundle


num_proc = mp.cpu_count()

def kmeans(points,k):
    kmeans = MiniBatchKMeans(n_clusters=k,random_state=0)
    sm = sparse.csr_matrix(points)
    return kmeans.fit_predict(sm)

def parallel_kmeans(fibers,points,ks):
    # selected_points = np.array(list(list(f[i] for f in fibers) for i in points))

    selected_points = np.swapaxes(fibers.reshape(-1,21,3),0,1)[points,:,:]

    pool = mp.Pool(num_proc)
    results = pool.starmap(kmeans,zip(selected_points,ks))
    pool.close()
    return results

def mapping(point_clusters,r):
    fibers_map = {}
    for i in range(r[0],r[1],1):
        key = "_".join(str(point_clusters[p][i]) for p in range(len(point_clusters)))
        if key not in fibers_map:
            fibers_map[key] = [i]
        else:
            fibers_map[key].append(i)
    return fibers_map

def merge_maps(maps):
    final_map = {}
    for map in maps:
        for key, val in map.items():
            if key not in final_map:
                final_map[key] = val
            else:
                final_map[key].extend(val)
    return final_map

def parallel_mapping(point_clusters,ranges):

    partial_mapping = partial(mapping,point_clusters)
    pool = mp.Pool(num_proc)
    results = pool.map(partial_mapping, ranges)
    pool.close()
    fiber_clusters = merge_maps(results)
    return fiber_clusters

def get_ranges(nfibers):
    total_len = nfibers
    len_groups, rest = divmod(total_len, num_proc)
    return list((n, min(n + len_groups, total_len)) for n in range(0, total_len, len_groups))

def is_inverted(f1,f2):
    dist_00 = np.linalg.norm(f1[0]-f2[0])
    dist_020 = np.linalg.norm(f1[0]-f2[20])
    if dist_00 > dist_020:
        return True
    else:
        return False

def align_fibers(bundle):
    new_fibers = [bundle[0]]
    f1 = bundle[0]
    for f2 in bundle[1:]:
        if is_inverted(f1,f2):
            new_fibers.append(f2[::-1])
        else:
            new_fibers.append(f2)
    return np.asarray(new_fibers)

def centroid_mean_align(bundle):
    fibers = align_fibers(bundle)
    return np.asarray(sum(fibers)/len(fibers))

def parallel_reassignment(fibers,fiber_clusters,central_index,thr):
    size_filter = 5
    large_clusters,small_clusters,large_indices,small_indices,large_centroids,small_centroids = [], [], [], [], [], []
    ###########
    fibers_idx = []
    small_idx = []
    ###########
    small_indices = []
    for key,indices in fiber_clusters.items():
        if len(indices) > size_filter:
            large_indices.append(int(key.split("_")[central_index]))
            c = [fibers[i] for i in indices]
            large_clusters.append(c)
    ###########
            fibers_idx.append(indices)
    ###########

            large_centroids.append(centroid_mean_align(c))
        else:
            small_indices.append(int(key.split("_")[central_index]))
            c = [fibers[i] for i in indices]
            small_clusters.append(c)
    ###########
            small_idx.append(indices)
    ###########
            small_centroids.append(centroid_mean_align(c))
    result = ffclust(21, np.asarray(small_centroids).ravel(), np.asarray(large_centroids).ravel(),
                    thr,len(small_centroids),len(large_centroids))
    reassignment = result[:len(small_centroids)]
    # reassignment = seg.segmentation(21,thr, large_centroids,small_centroids,len(small_centroids), len(large_centroids))
    count = 0
    num_fibers_reass = 0
    num_discarded = 0
    for small_index,large_index in enumerate(reassignment):
        fibers = small_clusters[small_index]
        fibers_idx_tmp = small_idx[small_index]

        if int(large_index)!=-1:
            large_clusters[large_index].extend(fibers)
    ###########
            fibers_idx[large_index].extend(fibers_idx_tmp)
    ###########
            num_fibers_reass += len(fibers)
            count+=1
        else:
            if len(fibers)>2:
                recover_cluster = small_clusters[small_index]
                large_clusters.append(recover_cluster)
    ###########
                fibers_idx.append(fibers_idx_tmp)
    ###########
                large_indices.append(small_indices[small_index])
            else:
                num_discarded +=1
    return large_clusters,large_indices, fibers_idx

def get_groups(clusters,central_index,fiber_index):
    groups = {}
    for i,cluster in enumerate(clusters):
        index = central_index[i]
        if index not in groups:
            groups[index] = [[cluster],[fiber_index[i]]]
        else:
            groups[index][0].append(cluster)
            groups[index][1].append(fiber_index[i])
    return groups

def matrix_dist(streamlines, get_max=True, get_mean=False):
    x = np.asarray(streamlines)
    distances = ((np.stack((x,x[:,::-1]))[:,None]-x[:,None])**2).sum(axis=4)
    if get_max and get_mean:
        max_distances = np.sqrt(distances.max(axis=3).min(axis=0))
        mean_distances = np.sqrt(distances.mean(axis=3).min(axis=0))
        return max_distances, mean_distances
    elif get_max:
        return np.sqrt(distances.max(axis=3).min(axis=0))
    elif get_mean:
        return  np.sqrt(distances.mean(axis=3).min(axis=0))
    else:
        print('error, should return atleast one matrix')

def create_graph(centroids,thr):
    matrix_dist_ = matrix_dist(centroids, get_max=True)
    matrix_dist_[matrix_dist_>thr] = 0
    G = nx.from_numpy_matrix(matrix_dist_)
    return G

def join(thr,group):
    centroids = [centroid_mean_align(c) for c in group[0]]
    graph = create_graph(centroids,thr)
    cliques = sorted(nx.find_cliques(graph), key=len, reverse=True)
    visited = {}
    new_clusters,new_centroids = [] ,[]
    new_clusters_fiber_idx = []
    for clique in cliques:
        new_cluster = []
        new_cluster_idx = []
        for node in clique:
            # print(node)
            if node not in visited:
                cluster = group[0][node]
                cluster_idx = group[1][node]
                new_cluster.extend(cluster)
                new_cluster_idx.extend(cluster_idx)
                visited[node] = True
        if len(new_cluster)>0:
            new_clusters.append(new_cluster)
            new_centroids.append(centroid_mean_align(new_cluster))
            ############################
            new_clusters_fiber_idx.append(new_cluster_idx)
            ############################
    # print(len(visited))
    return new_clusters,new_centroids,new_clusters_fiber_idx

def parallel_join(fiber_clusters,cluster_indices,fiber_indices,thr):
    groups = get_groups(fiber_clusters,cluster_indices,fiber_indices)
    # print('groups', groups)
    partial_join = partial(join,thr)
    pool = mp.Pool(num_proc)
    results = pool.map(partial_join, [g for key,g in groups.items()])
    # print('len de groups: ',len(groups) , ' len results: ', len(results))
    pool.close()
    new_clusters,new_centroids = [], []
    new_clusters_fiber_idx = []
    for clust,centroids,fiber_idx in results:
        new_clusters.extend(clust)
        new_clusters_fiber_idx.extend(fiber_idx)
        new_centroids.extend([c] for c in centroids)
    # print('len new_clusters:', len(new_clusters), '   ', len(new_centroids))
    return new_clusters,new_centroids,new_clusters_fiber_idx

class FFClustSegmentation(SegmentationHandler):
    # fibers = IO.read_bundles(args.infile)
    # clusters,centroids,log = clustering.fiber_clustering(fibers,args.points,args.ks,args.thr_seg,args.thr_join)



    def __init__(self, bundle, shaderDict):
        super().__init__(bundle, shaderDict)

        self.segmentationIdentifier = SegmentationTypes.FFClust

        self.fileName = 'FFclust' # temporal

        self.alpha = 0.8

        self.clusters = []

        # Default values:
        self.selected_points = list((0,3,10,17,20))             # Points to be used in map clustering
        self.ks = list((300, 200, 200, 200, 300))               # Number of clusters to be used for each point in K-Means for map
        self.thr_seg = 6                                                                # Minimum threshold for segmentation
        self.thr_join = 6                                                               # Minimum threshold for join

        self.colorTableTexture = None
        self.color = np.array([0.5,0.5,0.5,2.0], dtype=np.float32).reshape((1,4))
        self._loadColorTexture()

        self.configFiberValidator()

        # Creates VBOs & an EBO, then populates them with the point, normal and color data. The elements data goes into the EBO
        self._loadBuffers()
        self.buildVertex2Fiber()
        self.vboAndLinkVertex2Fiber()

        # Check if bundle has fixed point size per fiber (21 vertex per fiber)
        if (self.fiberSizes != 21).sum():
            print('Resampling...')
            self.fixedPoints = np.empty(self.curvescount*self.fiberLengthTimes3, dtype=np.float32)
            reSampleBundle(
                    self.points,
                    self.fiberSizes,
                    self.curvescount,
                    self.fixedPoints,
                    self.fiberLengthTimes3//3)

            print('Done.')
        else:
            self.fixedPoints = self.points

        self.fixedPoints = self.fixedPoints.reshape(self.curvescount,21,3)

        # self.segmentedBundles = []
        # self.validBundles = []

        self.boundingbox = BoundingBox(shaderDict, self, bundle.boundingbox.dims, bundle.boundingbox.center)
        print('Maximum number of clusters: ', np.iinfo(np.uint16).max)


    def segmentMethod(self):
        print('SEGMENTATION')
        point_clusters = parallel_kmeans(self.fixedPoints,self.selected_points,self.ks)
        fiber_clusters_map = parallel_mapping(point_clusters,get_ranges(self.curvescount))
        central_index = int((len(self.selected_points) - 1)/2)
        fiber_clusters,cluster_indices,fibers_idx = parallel_reassignment(self.fixedPoints,fiber_clusters_map,central_index,self.thr_seg)
        final_clusters,centroids,final_fibers_idx = parallel_join(fiber_clusters,cluster_indices,fibers_idx,self.thr_join)

        self.clusters = final_fibers_idx

        self.fiberValidator[:self.curvescount] = len(self.clusters)

        for index,fibers in enumerate(final_fibers_idx):
            self.fiberValidator[fibers] = index

        self.updateColor(len(final_fibers_idx))
        self.updateColorTexture()

        print(len(self.clusters))
        print(len(final_clusters))


    def updateFiberValidator(self):
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.validFiberTexture)

        GL.glTexSubImage2D(GL.GL_TEXTURE_2D, 0, 0, 0, *self.validFiberTexDims, GL.GL_RED_INTEGER, GL.GL_UNSIGNED_SHORT, self.fiberValidator)



    def defaultFiberValidator(self):
        return np.zeros(self.validFiberTexDims[1]*self.validFiberTexDims[0], dtype=np.uint16)


    def configFiberValidator(self):
        maxTexDim = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)

        if maxTexDim*maxTexDim < self.curvescount:
            raise ValueError('Fiber data set to big for GPU, need to change render mode...')
        self.validFiberTexDims = (maxTexDim, int(math.ceil(self.curvescount/maxTexDim)))
        # self. validFiberTexDims = findIntegersMultiplierFor(self.curvescount, maxTexDim)

        self.fiberValidator = self.defaultFiberValidator()

        self.validFiberTexture = GL.glGenTextures(1)

        GL.glActiveTexture(GL.GL_TEXTURE1)

        GL.glBindTexture(GL.GL_TEXTURE_2D, self.validFiberTexture)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)

        bgColor = np.array([1,1,1,1], dtype=np.float32)
        GL.glTexParameterfv(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_BORDER_COLOR, bgColor)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_R16UI, self.validFiberTexDims[0], self.validFiberTexDims[1], 0, GL.GL_RED_INTEGER, GL.GL_UNSIGNED_SHORT, self.fiberValidator)







    def _loadColorTexture(self):
        ''' It generates a texture (if not already). Then makes the texture0 the active one and loads the color table as a 1D texture.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''
        maxTexDim = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)

        if maxTexDim*maxTexDim < len(self.color):
            raise ValueError('To many bundles inside the atlas... Color texture not big enought.')
        self.validBundleColorTexDims = findIntegersMultiplierFor(len(self.color), maxTexDim)
        print(self.validBundleColorTexDims)

        if self.colorTableTexture == None:
            self.colorTableTexture = GL.glGenTextures(1)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_BORDER)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_BORDER)

        bgColor = np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float32)
        GL.glTexParameterfv(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_BORDER_COLOR, bgColor)

        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)

        # GL.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_RGBA, 1, 0, GL.GL_RGBA, GL.GL_FLOAT, self.atlas.colorTable.flatten())
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, *self.validBundleColorTexDims, 0, GL.GL_RGBA, GL.GL_FLOAT, self.color.flatten())


    def updateColor(self, n_of_clusters):
        self.color = np.empty((n_of_clusters+1,4), dtype=np.float32)
        self.color[-1] = [0.5,0.5,0.5,2.0]

        for i in range(n_of_clusters):
            self.color[i] = [random.random(), random.random(), random.random(), 2.0]


    def updateColorTexture(self):
        maxTexDim = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)

        if maxTexDim*maxTexDim < len(self.color):
            raise ValueError('To many bundles inside the atlas... Color texture not big enought.')
        self.validBundleColorTexDims = findIntegersMultiplierFor(len(self.color), maxTexDim)
        print(self.validBundleColorTexDims)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.colorTableTexture)

        # GL.glTexImage1D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, len(self.atlas.bundlesNames)+1, 0, GL.GL_RGBA, GL.GL_FLOAT, self.atlas.colorTable.flatten())
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, *self.validBundleColorTexDims, 0, GL.GL_RGBA, GL.GL_FLOAT, self.color.flatten())


    def _loadBuffers(self):
        ''' It generates a VAO, VBO and EBO.
        It populates them with the vertex, normal and colors (the VBO), also the elements (the EBO).

        It finishes when the buffers are linked with attributes on the vertex shader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        # Reference for vbos, ebos and attribute ptrs
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        # self.ebo = GL.glGenBuffers(1)

        # VBO positions
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[0])

        positionAttribute =     self.shader[0].attributeLocation('vertexPos')
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute,3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        # VBO normals
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo[1])

        normalAttribute =       self.shader[0].attributeLocation('vertexNor')
        GL.glEnableVertexAttribArray(normalAttribute)
        GL.glVertexAttribPointer(normalAttribute,       3, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)


    def loadUniforms(self):
        super().loadUniforms()

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('notAssigned'), len(self.clusters))

        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('colorTable'), 0)
        GL.glUniform1i(self.shader[self.selectedShader].glGetUniformLocation('fiberValidator'), 1)



from pathlib import Path
from json import load

def load_config(file="config.json"):
    p = Path(file)

    if p.is_absolute():
        if not p.exists():
            raise FileNotFoundError(f"Config file {p} not found.")
    else:
        cwd = Path.cwd()
        if (cwd / p).exists():
            p = cwd / p
        elif (cwd / "test" / p).exists():
            p = (cwd / "test" / p)
        elif (Path(__file__).parent / p).exists():
            p = (Path(__file__).parent / p)
        else:
            raise FileNotFoundError(f"Config file {p} not found.")
    with open(str(p), 'rt') as config_file:
        return load(config_file)


def try_import() -> bool:
    try:
        import phybers
        return True
    except ImportError:
        return False


def ffclust():
    from phybers.clustering import ffclust
    try:
        ffclust(*load_config()["ffclust"])
        return True
    except SystemError:
        return False


def hclust():
    from phybers.clustering import hclust
    try:
        hclust(*load_config()["hclust"])
        return True
    except SystemError:
        return False


def segmentation():
    from phybers.segment import fiberseg
    try:
        fiberseg(*load_config()["segmentation"])
        return True
    except SystemError:
        return False


def deform():
    from phybers.utils import deform
    try:
        deform(*load_config()["deform"])
        return True
    except SystemError:
        return False


def intersection():
    from phybers.utils import intersection
    try:
        intersection(*load_config()["intersection"])
        return True
    except SystemError:
        return False


def sampling():
    from phybers.utils import sampling
    try:
        sampling(*load_config()["sampling"])
        return True
    except SystemError:
        return False


def postprocessing():
    from phybers.utils import postprocessing
    try:
        postprocessing(*load_config()["postprocessing"])
        return True
    except SystemError:
        return False


def test_import_succes():
    print(Path.cwd())
    assert try_import()


def test_ffclust():
    assert ffclust()


def test_hclust():
    assert hclust()


def test_segmentation():
    assert segmentation()


def test_deform():
    assert deform()


def test_intersection():
    assert intersection()


def test_sampling():
    assert sampling()


def test_postprocessing():
    assert postprocessing()



# -- coding: utf-8 --
"""
FFclust Sub-Module
"""
import logging
import time
import os
import numpy as np
import joblib
from ...utils import read_bundle, sampling
from .clustering import split_fibers, parallel_points_clustering, MapClustering,\
    small_clusters_reassignment, get_groups, parallel_group_join_clique
from .utils import clusters_fibers21p, save_clusters, save_clusters_centroids

def ffclust(file_in: str, dir_out: str, assign_thr: int=6, join_thr: int=6) -> bool:
    """
    Intra-subject clustering algorithm aims to identify compact and homogeneous fiber clusters on a large tractography dataset.

    Parameters
    ----------
    file_in : str
        Tractography data file in *'.bundles'* format.
    dir_out : str
        Directory to store all the results generated by the algorithm.
    assign_thr : str
        Maximum distance threshold for the cluster reassignment in *mm*, default: *6*.
    join_thr : str
        Maximum distance threshold for the cluster merge in *mm*, default: *6*.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Clusters : bundles files
        Directory that stores all the fiber clusters found in different '.bundles' files.
        The file names are labeled with integer numbers ranging from zero to the total number of fiber clusters found.
    Centroids : bundles file
        Directory that contains the centroid for each created cluster in same *'.bundles'* files.
        The firt fiber of bundle centroid is corresponding with centroid calculated for cluster one an so on.
    Index of fibers per clusters : text file
        Text file that stores the fiber index input for each of the detected clusters. The fiber indexes are extracted from the tractography input.
        The first line corresponds to cluster zero and so on.

    Examples
    --------
    To test `ffclust()`,  download the data from the `link ffclust  <https://www.dropbox.com/sh/3z7uriisict4xxn/AADCsGDWEPeoHUVm4FqHzLdaa?dl=1>`_ .
    Then, open a Python terminal and run the following commands.

    >>> from phybers.clustering import ffclust
    >>> ffclust (file_in ='fibers_test.bundles', dir_out = 'ffclust_result', assign_thr = 6, join_thr = 6)

    You will locate the clustering results in the 'ffclust_result' directory.
    """

    if isinstance(assign_thr, str):
        assign_thr = int(assign_thr)
    if isinstance(join_thr, str):
        join_thr = int(join_thr)

    points = np.array([0,3,10,17,20])
    ks = np.array([200,300,300,300,200])

    os.makedirs(dir_out, exist_ok=True)

    final_bundles21p_dir = os.path.join(dir_out, 'FinalBundles21p')
    os.makedirs(final_bundles21p_dir, exist_ok=True)

    final_centroids_dir = os.path.join(dir_out,'FinalCentroids')
    os.makedirs(final_centroids_dir, exist_ok=True)

    object_dir = os.path.join(dir_out, 'output')
    os.makedirs(object_dir, exist_ok=True)
    data=read_bundle(file_in)

    npf=0
    for i in range(len(data)-1):
        if len(data[i]) != len(data[i+1]):
            npf=21
            break


    if npf == 21:
        fibers21p = os.path.join(object_dir,'fiberorig_21p.bundles')
        sampling(file_in, fibers21p, 21)
        fibers = np.asarray(read_bundle(fibers21p))
    else:
        fibers = np.asarray(data)

    # New logging file each time, changing filemode to a should append instead.
    logging.basicConfig(filename=dir_out+'/info.log', filemode='w', level=logging.INFO)

    t1 = time.time()
    X = split_fibers(fibers[:,points,:],points)
    labels, clusterers = parallel_points_clustering(X=X, ks=ks)
    logging.info('Tiempo Kmeans: {}'.format(time.time() - t1))

    t1 = time.time()
    m = MapClustering()
    map_clusters = m.cluster(fibers, labels)
    logging.info('Tiempo Map: {}'.format(time.time() - t1))

    map_output_filename = os.path.join(object_dir, 'clusters_map.txt')
    save_clusters(dataset=fibers, clusters=map_clusters, filename=map_output_filename)

    with open(dir_out+'/stats.txt', 'w') as f:
        f.write('Number of clusters in map_clusters: {}\n'.format(len(map_clusters)))
        f.write('Number of fibers in map_clusters: {}\n'.format(sum(map_clusters.clusters_sizes())))

    for p,k, clusterer in zip(points, ks, clusterers):
        joblib.dump(clusterer, object_dir + '/clusterer-Point{}-k{}.pkl'.format(p,k))

    """Write .bundles .bundlesdata of small (< 1 fiber) and long centroids (> 1 fiber) of clusters
    and reassign small clasters to large clusters. Input:
    ouput: path of output file directory to write the clusters
    min_size_filter = minimum size to get the largest clusters
    max_size_filter = maximum size to get the clusters with only 1 fiber
    input_dir = path of the input dir that contains the results of segmetation/reassignation
    threshold = minimum distance for the segmentation's method
    """
    # Compenzando la reasignación.
    t1 = time.time()
    actual_clusters = small_clusters_reassignment(clusters=map_clusters,
                                                  min_size_filter=6,
                                                  max_size_filter=5,
                                                  input_dir = 'segmentation/bundles/result/parallelFastCPU',
                                                  threshold = assign_thr,
                                                  refdata = fibers)

    ident_clusters = dict()
    for i,c in enumerate(actual_clusters):
        ident_clusters[str(c.indices)] = i

    logging.info('Tiempo Segmentacion {}'.format(time.time() - t1))

    t1 = time.time()
    point_index = 10
    ngroups = ks[len(ks)//2]

    centroids = np.asarray([x.centroid for x in actual_clusters])
    centroids_points = centroids[:,point_index]
    clusterer=clusterers[len(points)//2]
    labels = clusterer.predict(centroids_points)

    groups = get_groups(labels, ngroups=ngroups)

    final_clusters = parallel_group_join_clique(actual_clusters, groups, fibers, final_bundles21p_dir, final_centroids_dir, ident_clusters, object_dir, join_thr)
    final_centroids_filename = os.path.join(object_dir, 'final_centroids.txt')
    save_clusters_centroids(clusters=final_clusters, filename=final_centroids_filename)

    final_bundles_dir = os.path.join(dir_out, 'FinalBundles')
    if npf == 21:
        os.makedirs(final_bundles_dir, exist_ok=True)
        clusters_fibers21p(final_clusters, final_bundles_dir, data)
    else:
        os.rename(final_bundles21p_dir, final_bundles_dir)
    return True



'''
InPlaceSegmentationDialog.exportBundle ----> trk not implemented
'''

from .GLContext import GLWidget
from PyQt5 import QtGui, QtWidgets, uic, QtCore
from .ui.Controllers.VisDialog import VisDialog
from .Framework.Tools.performance import timeit
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_viewer = files('phybers.fibervis') / 'ui' / 'viewer.ui'

class WindowController(QtWidgets.QMainWindow):
    ''' Window controller for the app. It has function for several action.
    '''
    def __init__(self):
        super().__init__()

        glFormat = QtGui.QSurfaceFormat()
        glFormat.setVersion(4, 1)
        glFormat.setProfile(QtGui.QSurfaceFormat.CoreProfile)
        glFormat.setSamples(4)

        QtGui.QSurfaceFormat.setDefaultFormat(glFormat)

        # versionProfile = QtGui.QOpenGLVersionProfile()
        # versionProfile.setVersion(4, 1)
        # versionProfile.setProfile(QtGui.QSurfaceFormat.CoreProfile)

        # self.contextHandler = GLContext.GLWidget(versionProfile, self)
        self.contextHandler = GLWidget(self)
        self.contextHandler.setFocusPolicy(QtCore.Qt.StrongFocus)

        self.ui = uic.loadUi(_viewer, self)
        self.visObj = VisDialog(self)

        self.ui.setCentralWidget(self.contextHandler)

        self.ui.setWindowTitle('leFiber v1.0.0')

        # Connection of GUI to GLCBontext
        self.ui.actionOpenBundle.triggered.connect(self.openBundleFile)
        self.ui.actionOpenMesh.triggered.connect(self.openMeshFile)
        self.ui.actionOpenMRI.triggered.connect(self.openMRIFile)
        self.ui.actionAddROISphere.triggered.connect(self.createROI)

        #Toggle Bounding Boxes
        self.ui.actionBoundingBoxes.changed.connect(
                lambda: self.contextHandler.set_draw_bbox(self.ui.actionBoundingBoxes.isChecked()))
        self.ui.actionBoundingBoxes.setChecked(True)

        # Visualization objects window
        # changed
        # hovered
        # toggled
        self.ui.actionVisualizationObjects.changed.connect(self.visObj.toggleVisWin(self.ui.actionVisualizationObjects.isChecked, self.pos, self.size))

        # Closed visualization win
        self.visObj.closed.connect(self.unCheckActionVisualizationObjects)

        # New visualization object
        self.contextHandler.updateViewingObjects.connect(self.visObj.updateTree)
        self.contextHandler.updateSettingWindow.connect(self.visObj.updateSettingWindow)

        # Modify visualization object
        self.visObj.changeObject.connect(self.contextHandler.modifyObject)
        self.visObj.modifySegmentation.connect(self.contextHandler.modifySegmentation)
        self.visObj.selectedObject.connect(self.contextHandler.selectedObject)

        # Notify R, T and S on objects
        self.contextHandler.objectsChanged.connect(self.visObj.objectsChanged)

        # We set the reference for the visualization objects
        self.visObj.bundleReference = self.contextHandler.bundles
        self.visObj.meshReference = self.contextHandler.meshes
        self.visObj.mriReference = self.contextHandler.mris
        self.visObj.roisReference = self.contextHandler.rois

        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if (event.mimeData().hasUrls()):
            event.acceptProposedAction()

    @timeit
    def dropEvent(self, event):
        for file in event.mimeData().urls():
            fileName = file.toLocalFile()

            extension = fileName.split('.')[-1]

            # Bundle
            if extension in self.contextHandler.validBundleExtension():
                self.contextHandler.addBundleFile([fileName])
            elif extension in self.contextHandler.validMeshExtension():
                self.contextHandler.addMeshFile([fileName])
            elif extension in self.contextHandler.validMRIExtension():
                self.contextHandler.addMRIFile([fileName])
            # elif extension == "bundlesdata":
            #       self.contextHandler.addBundleFile([fileName[:-4]])
            else:
                print('File extension not sopported: ', fileName)


    def closeEvent(self, event):
        print('Cleaning...')
        self.contextHandler.cleanOpenGL()

        super().closeEvent(event)


    def keyPressEvent(self, event):
        # print('Key press: {0}'.format(event.key()))
        pass
        # self.contextHandler.setWindowState(QtCore.Qt.WindowFullScreen)
        # self.ui.showFullScreen()
        # if self.tFullScreen and event.key() == QtCore.Qt.Key_Escape:
        #       event.accept()
        #       self.toggleFullScreen()
        # else:
        #       event.ignore()


    @QtCore.pyqtSlot()
    def unCheckActionVisualizationObjects(self):
        self.ui.actionVisualizationObjects.setChecked(False)


    @QtCore.pyqtSlot()
    def openBundleFile(self):
        ''' Function that gets called when pressing the open bundle file action
        '''

        # Gets direction of bundle files
        formats = "" # *.bundles *.trk *.tck
        for f in self.contextHandler.validBundleExtension():
            formats += "*."+f+" "
        bundleFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select Bundle", "", "Bundles ("+formats+")")

        if len(bundleFiles) == 0:
            return

        self.contextHandler.addBundleFile(bundleFiles)


    @QtCore.pyqtSlot()
    def openMeshFile(self):
        ''' Function that gets called when pressing the open mesh file action
        '''

        # Gets direction of bundle files
        formats = ""
        for f in self.contextHandler.validMeshExtension():
            formats += "*."+f+" "
        meshFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select Mesh", "", "Mesh ("+formats+")")

        if len(meshFiles) == 0:
            return

        self.contextHandler.addMeshFile(meshFiles)


    @QtCore.pyqtSlot()
    def openMRIFile(self):
        ''' Function that gets called when pressing the open MRI file action
        '''

        # Gets direction of bundle files
        formats = ""
        for f in self.contextHandler.validMRIExtension():
            formats += "*."+f+" "
        mriFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select MRI", "", "MRI ("+formats+")")

        if len(mriFiles) == 0:
            return

        self.contextHandler.addMRIFile(mriFiles)


    @QtCore.pyqtSlot()
    def createROI(self):
        '''
        '''

        self.contextHandler.addROI()


    def testSlot(self):
        print('testSlot!')



from collections import deque
from enum import Enum
import numpy as np


class TokenType(Enum):
    NUMBER = 0
    OPERATOR = 1
    LEFTBRACKET = 2
    RIGHTBRACKET = 3


class Associative(Enum):
    LEFT = 0
    RIGHT = 1


class Token:
    value = None
    tokenType = None

    def __init__(self, value=None):
        if value is None:
            return

        self.value = value

        if isinstance(value, str) and len(value) == 1:
            if value == '(':
                self.tokenType = TokenType.LEFTBRACKET
            elif value == ')':
                self.tokenType = TokenType.RIGHTBRACKET
            else:
                self.tokenType = TokenType.OPERATOR
        else:
            self.tokenType = TokenType.NUMBER


class LogicalMath:

    MathTable = {'&' : {'func' : np.logical_and, 'nOfOperand' : 2},
                            '|' : {'func' : np.logical_or, 'nOfOperand' : 2},
                            '!' : {'func' : np.logical_not, 'nOfOperand' : 1},
                            '^' : {'func' : np.logical_xor, 'nOfOperand' : 2}}

    PrecedenceTable = {     '|' : 0,
                                            '^' : 0,
                                            '&' : 1,
                                            '!' : 2}

    AssociativeTable = {'&' : Associative.LEFT,
                                            '|' : Associative.LEFT,
                                            '!' : Associative.LEFT,
                                            '^' : Associative.LEFT}


class Math:
    MathTable = {   '^' : {'func' : lambda a,b : pow(a,b), 'nOfOperand' : 2},
                                    '*' : {'func' : lambda a,b : a*b, 'nOfOperand' : 2},
                                    '/' : {'func' : lambda a,b : a/b, 'nOfOperand' : 2},
                                    '+' : {'func' : lambda a,b : a+b, 'nOfOperand' : 2},
                                    '-' : {'func' : lambda a,b : a-b, 'nOfOperand' : 2}}

    PrecedenceTable = {     '^' : 4,
                                            '*' : 3,
                                            '/' : 3,
                                            '+' : 2,
                                            '-' : 2}

    AssociativeTable = {'(', Associative.LEFT,
                                            '^', Associative.RIGHT,
                                            '*', Associative.LEFT,
                                            '/', Associative.LEFT,
                                            '+', Associative.LEFT,
                                            '-', Associative.LEFT}


def infix2postfix(tokens, precedenceTable):
    if not isinstance(tokens, deque):
        print('shunting_yard: infix2postfix input is not a deque... casting into one.')
        tokens = deque(tokens)

    operatorStack = []
    outputqueue = deque()

    while len(tokens):
        token = tokens.popleft()

        if token.tokenType == TokenType.NUMBER:
            outputqueue.append(token)

        # elif token in functionTable:
            # operatorStack.append(token)

        elif token.tokenType == TokenType.OPERATOR:
            while ( len(operatorStack) and
                    operatorStack[-1].tokenType != TokenType.LEFTBRACKET and
                    precedenceTable[token.value] < precedenceTable[operatorStack[-1].value]):

                outputqueue.append(operatorStack.pop())
            operatorStack.append(token)

        elif token.tokenType == TokenType.LEFTBRACKET:
            operatorStack.append(token)

        elif token.tokenType == TokenType.RIGHTBRACKET:
            while len(operatorStack) and operatorStack[-1].tokenType != TokenType.LEFTBRACKET:
                outputqueue.append(operatorStack.pop())

            if not len(operatorStack):
                return -1

            operatorStack.pop()

    while len(operatorStack):
        outputqueue.append(operatorStack.pop())
        if outputqueue[-1].tokenType == TokenType.LEFTBRACKET:
            return -1

    return outputqueue


def postfixEvaluator(tokens, MathTable):
    if not isinstance(tokens, deque):
        print('shunting_yard: postfixEvaluator input is not a deque... casting into one.')
        tokens = deque(tokens)

    stack = []

    while len(tokens):
        token = tokens.popleft()

        if token.tokenType == TokenType.NUMBER:
            stack.append(token)

        else:
            right_operand = stack.pop()

            if MathTable[token.value]['nOfOperand'] == 1:
                stack.append(Token(MathTable[token.value]['func'](right_operand.value)))
            else:
                left_operand = stack.pop()
                stack.append(Token(MathTable[token.value]['func'](left_operand.value, right_operand.value)))

    if len(stack) == 1:
        return stack.pop()

    else:
        raise ValueError


def shuntingYard(infix, operation='Logical'):
    if operation == 'Logical':
        table = LogicalMath
    elif operation == 'Math':
        table = Math
    else:
        raise ValueError('Not recogniced.')

    infixTokens = deque([Token(i) for i in infix])
    postfixTokens = infix2postfix(infixTokens, table.PrecedenceTable)

    if postfixTokens == -1:
        raise ValueError('shunting_yard: Bracket not paired.')

    # print([i.value for i in postfixTokens])

    evaluation = postfixEvaluator(postfixTokens, table.MathTable)

    return evaluation.value



import functools
import sklearn.cluster
import numpy as np
import os
from dipy.segment.clustering import ClusterCentroid, \
                                    ClusterMapCentroid, \
                                    Clustering
from joblib import Parallel, delayed
from ...utils import write_bundle
from . import metric as metric
from . import utils as utils
from .c_wrappers import seg as segmentation


def cluster_kmeans(x, k, random_state=0):
    kmeans = sklearn.cluster.MiniBatchKMeans(n_clusters=k, n_init=3,
                                             random_state=random_state)
    return kmeans.fit_predict(x), kmeans


def parallel_points_clustering(X, ks, n_jobs=-1, backend='threading'):
    """
    Compute MiniBatchKMeans for each point according to cluster size
    Do in parallel, conserve clusterers object.
    """
    results = Parallel(n_jobs=n_jobs, backend=backend)(delayed(cluster_kmeans)(x, k) for x, k in zip(X, ks))
    labels, clusterers = list(map(list, zip(*results)))
    labels = np.array(labels).T
    return labels, clusterers


def map_clustering(labels):
    """
    Create groups based on the labels that composed each element

    @arg labels: List of lists, each element in the dataset
    has m features, then
    if each feature has an associated label each element is the
    collection of such labels, in this case the labels come from
    a previous clustering step performed in each feature.
    """

    labels_it = map(tuple, labels)
    clusters = []
    d = {}

    # For every element check its label
    # if its new then create a new cluster
    # else assign to the cluster that has all elements with the
    # same label
    for i, label in enumerate(labels_it):
        if label in d:
            clusters[d[label]].append(i)
        else:
            clusters.append([i])
            d[label] = len(d)
    return clusters


def split_fibers(fibers, points):
    return np.array(np.split(fibers, len(points), axis=1))[:, :, 0]


# Wrapper for dipy given that we have clusters and not centroids.
def clusters_to_clustermap(clusters, fibers, ids=None):
    """
    Wrapper for dipy clustering interface
    @arg clusters : list of lists representing the clusters, each
    inner list contains the index of the elements in the dataset that
    conform this cluster.

    @arg fibers : dataset
    """

    cluster_map = ClusterMapCentroid(fibers)
    if ids:
        for i, cluster in zip(ids, clusters):
            # Compute a centroid as a mean of its elements.

            centr = fibers[cluster].mean(axis=0)
            c = ClusterCentroid(centroid=centr,
                                id=i, indices=cluster,
                                refdata=fibers)
            cluster_map.add_cluster(c)
        return cluster_map
    for i, cluster in enumerate(clusters):
        # Compute a centroid as a mean of its elements.
        centr = fibers[cluster].mean(axis=0)
        c = ClusterCentroid(centroid=centr,
                            id=i, indices=cluster,
                            refdata=fibers)
        cluster_map.add_cluster(c)
    return cluster_map


def get_groups(labels, ngroups):
    groups = [[] for i in range(ngroups)]
    for i, l in enumerate(labels):
        groups[l].append(i)
    return groups


def small_clusters_reassignment(clusters,
                                min_size_filter, max_size_filter,
                                input_dir, threshold, refdata):
    """
    Write .bundles .bundlesdata of small (< 1 fiber) and long centroids (> 1 fiber) of clusters.
    """
    small_clusters = clusters.get_small_clusters(max_size_filter)
    large_clusters = clusters.get_large_clusters(min_size_filter)
    small_centroids = np.asarray([x.centroid for x in small_clusters])
    large_centroids = np.asarray([x.centroid for x in large_clusters])
    reassignment = segmentation(21, threshold, large_centroids, small_centroids, len(small_centroids), len(large_centroids))

    count = 0
    num_fibers_reass = 0
    num_discarded = 0
    # Reassign small clusters to large clusters.
    for small_index,large_index in enumerate(reassignment):
        fibers = small_clusters[small_index].indices
        if int(large_index)!=-1:
            # Hacer bucle para reasignar todas las fibras.
            for fiber in fibers:
                centroid = large_clusters[large_index].centroid
                f = small_clusters[small_index].refdata[fiber]
                large_clusters[large_index].assign(id_datum=fiber, features= small_clusters[small_index].refdata[fiber])
                num_fibers_reass += 1
            metric.recalc_centroid(large_clusters[large_index].indices, refdata)
            count+=1
        else:
            if len(fibers)>2:
                recover_cluster = small_clusters[small_index]
                large_clusters.append(recover_cluster)
            else:
                num_discarded +=1
    return large_clusters

# Classes for use with dipy ClusteringMap objects.
class MapClustering(Clustering):
    # Must return ClusterMap object
    # So we wrap our function around this
    def cluster(self, fibers, labels):
        clusters = map_clustering(labels)
        return clusters_to_clustermap(clusters, fibers)

def joinable_clusters(clique, visited):
    """
    Visits the nodes of a clique and returns nodes that are not visited.
    In the clustering, the nodes represent the centroids of the clusters.
    """
    clusters_ids = []
    for node in clique:
        # Some clusters may be joined already.
        if not visited[node]:
            clusters_ids.append(node)
            visited[node] = True
    return clusters_ids

def clique_join(clusters, refdata, joined_bundles_dir, ident_clusters, object_dir, threshold):
    """
    Returns a list of new clusters(dipy.clustering.ClusterCentroid container)
    based on computing the distance between centroids and using this matrix as input
    to finding cliques, distances greater than the specified threshold are ignored.
    """

    try:
        from networkx import from_numpy_array, find_cliques # type: ignore
    except ImportError:
        from networkx import from_numpy_matrix as from_numpy_array, find_cliques # type: ignore
    centroids = np.asarray([x.centroid for x in clusters])

    max_dists = metric.matrix_dist(centroids, get_max=True)

    max_dists[max_dists > threshold] = 0 # type: ignore
    network = from_numpy_array(max_dists)
    cliques = sorted(find_cliques(network), key=len, reverse=True)

    visited = np.zeros(len(centroids), np.bool_)
    new_clusters = []
    cluster_names = []
    map_cluster_names = {}
    for clique in cliques:
        # Each clique is potentially a cluster.
        clusters_ids = joinable_clusters(clique, visited)
        if not clusters_ids:
            continue
        # Need to copy indices to preserve number of streamlines, no idea why.
        ob = ClusterCentroid(centroid=clusters[clusters_ids[0]].centroid,
                             indices=clusters[clusters_ids[0]].indices[:],
                             refdata=clusters[clusters_ids[0]].refdata)
        # Node value indexes into centroids.
        if (len(clusters_ids)) > 1:
            for idx in clusters_ids[1:]:
                for index in clusters[idx].indices:
                    ob.assign(id_datum=index, features=refdata[index])
            ob.centroid = metric.recalc_centroid(ob.indices,refdata)
        new_clusters.append(ob)

        string_ids = ""
        for idx in clusters_ids:
            index_cluster = ident_clusters[str(clusters[idx].indices)]
            string_ids = string_ids + str(index_cluster)+"c"
        cluster_names.append(string_ids)
    for i,cluster in enumerate(new_clusters):
        map_cluster_names[cluster_names[i]] = cluster
    return map_cluster_names

def parallel_group_join_clique(clusters, groups, refdata,joined_bundles_dir,dir_out,final_centroids_dir,ident_clusters,object_dir,thr_join):
    """
    Same as group_join_clique, using multiprocessing.dummy.Pool for threading
    Return clusters after distributing them in groups and joining with both distances and
    cliques inside those groups.

    groups: Each group if a list of indices of clusters which are in that group.
    """
    import multiprocessing.dummy

    final_clusters = ClusterMapCentroid(refdata)
    ind = 0
    pool_input = []
    for group in groups:
        cls = [clusters[i] for i in group]
        pool_input.append(cls)

    p = multiprocessing.dummy.Pool()
    func=functools.partial(clique_join, refdata=refdata, joined_bundles_dir=joined_bundles_dir,
                           ident_clusters=ident_clusters, object_dir=object_dir, threshold=thr_join)
    results = p.map(func, pool_input)

    final_output_filename = os.path.join(dir_out, "bundles_id.txt")
    utils.save_join_clusters_fibers(results=results, filename=final_output_filename)

    final_centroids = []
    i = 0
    for cluster_names in results:
        for name in cluster_names:
            cluster = cluster_names[name]
            join_bundles_file = os.path.join(joined_bundles_dir, str(i)+".bundles")
            i+=1
            c = np.asarray(cluster[:])
            final_centroids.append(cluster.centroid)
            write_bundle(join_bundles_file,c)

    write_bundle(os.path.join(final_centroids_dir, "centroids.bundles"),np.asarray(final_centroids))

    for new_clusters in results:
        for k in new_clusters:
            new_clusters[k].id = ind
            ind += 1
            final_clusters.add_cluster(new_clusters[k])

    return final_clusters
